# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import numpy as np
import pytest


@pytest.mark.parametrize(
    "request_",
    [
        [],
        [
            {"name": "a", "func": "l_inf_norm"},
            {"name": "b", "func": "l_1_norm"},
            {"name": "c", "func": "l_2_norm"},
            {"name": "d", "func": "paired_cosine_similarities"},
            {"name": "e", "func": "paired_euclidean_distances"},
            {"name": "f", "func": "paired_manhattan_distances"},
            {"name": "g", "func": "paired_wasserstein_distances"},
        ],
        [{"name": "a", "func": "l_inf_norm"}, {"name": "b", "func": "l_1_norm"}],
        [{"name": "c", "func": "l_1_norm"}, {"name": "d", "func": "l_inf_norm"}],
        [{"name": "e", "func": "paired_wasserstein_distances"}],
    ],
)
@pytest.mark.parametrize(
    ("y_true", "y_pred"),
    [
        ([[1, 2, 3], [3, 4, 5]], [[-2, 4, 5], [-9, 0, 1]]),
        ([[1, 2], [3, 4], [4, 5]], [[-2, 4], [5, -9], [0, 1]]),
        ([1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1]),
        ([10], [1]),
    ],
)
def test_get_distance_metric_list(request_, y_true, y_pred) -> None:
    from dioptra_builtins.metrics.distance import get_distance_metric_list

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)

    result: list = get_distance_metric_list(request_)
    for res in result:
        assert isinstance(res[0], str)
        dists = res[1](y_true, y_pred)
        assert all([isinstance(dist, float) for dist in dists])


@pytest.mark.parametrize(
    "func",
    [
        "l_inf_norm",
        "l_1_norm",
        "l_2_norm",
        "paired_cosine_similarities",
        "paired_euclidean_distances",
        "paired_manhattan_distances",
        "paired_wasserstein_distances",
    ],
)
@pytest.mark.parametrize(
    ("y_true", "y_pred"),
    [
        ([[1, 2, 3], [3, 4, 5]], [[-2, 4, 5], [-9, 0, 1]]),
        ([[1, 2], [3, 4], [4, 5]], [[-2, 4], [5, -9], [0, 1]]),
        ([1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1]),
        ([10], [1]),
    ],
)
def test_get_distance_metric(func, y_true, y_pred) -> None:
    from dioptra_builtins.metrics.distance import get_distance_metric

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)

    metric = get_distance_metric(func)
    results = metric(y_true, y_pred)
    assert all([isinstance(res, float) for res in results])


@pytest.mark.parametrize(
    ("y_true", "y_pred", "expected"),
    [
        ([[1, 2, 3], [3, 4, 5]], [[-2, 4, 5], [-9, 0, 1]], [3.0, 12.0]),
        ([[1, 2], [3, 4], [4, 5]], [[-2, 4], [5, -9], [0, 1]], [3.0, 13.0, 4.0]),
        ([1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1], [3.0, 2.0, 2.0, 13.0, 5.0, 5.0]),
        ([10], [1], [9.0]),
    ],
)
def test_l_inf_norm(y_true, y_pred, expected) -> None:
    from dioptra_builtins.metrics.distance import l_inf_norm

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    expected = np.array(expected)
    result: np.ndarray = l_inf_norm(y_true, y_pred)

    assert all([isinstance(res, float) for res in result])
    assert np.allclose(expected, result)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "expected"),
    [
        ([[1, 2, 3], [3, 4, 5]], [[-2, 4, 5], [-9, 0, 1]], [7.0, 20.0]),
        ([[1, 2], [3, 4], [4, 5]], [[-2, 4], [5, -9], [0, 1]], [5.0, 15.0, 8.0]),
        ([1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1], [3.0, 2.0, 2.0, 13.0, 5.0, 5.0]),
        ([10], [1], [9.0]),
    ],
)
def test_l_1_norm(y_true, y_pred, expected) -> None:
    from dioptra_builtins.metrics.distance import l_1_norm

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    expected = np.array(expected)

    result: np.ndarray = l_1_norm(y_true, y_pred)

    assert all([isinstance(res, float) for res in result])
    assert np.allclose(expected, result)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "expected"),
    [
        ([[1, 2, 3], [3, 4, 5]], [[-2, 4, 5], [-9, 0, 1]], [4.123105625, 13.2664991]),
        (
            [[1, 2], [3, 4], [4, 5]],
            [[-2, 4], [5, -9], [0, 1]],
            [3.605551275, 13.152946437, 5.656854249],
        ),
        ([1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1], [3.0, 2.0, 2.0, 13.0, 5.0, 5.0]),
        ([10], [1], [9.0]),
    ],
)
def test_l_2_norm(y_true, y_pred, expected) -> None:
    from dioptra_builtins.metrics.distance import l_2_norm

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    expected = np.array(expected)

    result: np.ndarray = l_2_norm(y_true, y_pred)

    assert all([isinstance(res, float) for res in result])
    assert np.allclose(expected, result)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "expected"),
    [
        (
            [[1, 2, 3], [3, 4, 5]],
            [[-2, 4, 5], [-9, 0, 1]],
            [0.8366600265340757, -0.3435822761549333],
        ),
        (
            [[1, 2], [3, 4], [4, 5]],
            [[-2, 4], [5, -9], [0, 1]],
            [0.6, -0.40794006219005097, 0.7808688094430304],
        ),
        ([1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1], [-1.0, 1.0, 1.0, -1.0, np.nan, 1.0]),
        ([10], [1], [1.0]),
    ],
)
def test_paired_cosine_similarities(y_true, y_pred, expected) -> None:
    from dioptra_builtins.metrics.distance import (
        _flatten_batch,
        _normalize_batch,
        paired_cosine_similarities,
    )

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    expected = np.array(expected)

    result: np.ndarray = paired_cosine_similarities(y_true, y_pred)

    assert all([isinstance(res, float) for res in result])
    assert np.allclose(expected, result, equal_nan=True)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "expected"),
    [
        (
            [[1, 2, 3], [3, 4, 5]],
            [[-2, 4, 5], [-9, 0, 1]],
            [4.123105625617661, 13.2664991614216],
        ),
        (
            [[1, 2], [3, 4], [4, 5]],
            [[-2, 4], [5, -9], [0, 1]],
            [3.605551275463989, 13.152946437965905, 5.656854249492381],
        ),
        ([1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1], [3.0, 2.0, 2.0, 13.0, 5.0, 5.0]),
        ([10], [1], [9.0]),
    ],
)
def test_paired_euclidean_distances(y_true, y_pred, expected) -> None:
    from dioptra_builtins.metrics.distance import (
        _flatten_batch,
        paired_euclidean_distances,
    )

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    expected = np.array(expected)
    result: np.ndarray = paired_euclidean_distances(y_true, y_pred)

    assert all([isinstance(res, float) for res in result])
    assert np.allclose(expected, result)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "expected"),
    [
        ([[1, 2, 3], [3, 4, 5]], [[-2, 4, 5], [-9, 0, 1]], [7.0, 20.0]),
        ([[1, 2], [3, 4], [4, 5]], [[-2, 4], [5, -9], [0, 1]], [5.0, 15.0, 8.0]),
        ([1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1], [3.0, 2.0, 2.0, 13.0, 5.0, 5.0]),
        ([10], [1], [9.0]),
    ],
)
def test_paired_manhattan_distances(y_true, y_pred, expected) -> None:
    from dioptra_builtins.metrics.distance import (
        _flatten_batch,
        paired_manhattan_distances,
    )

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    expected = np.array(expected)
    result: np.ndarray = paired_manhattan_distances(y_true, y_pred)

    assert all([isinstance(res, float) for res in result])
    assert np.allclose(expected, result)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "expected"),
    [
        ([[1, 2, 3], [3, 4, 5]], [[-2, 4, 5], [-9, 0, 1]], [2.33333333, 6.6666666]),
        ([[1, 2], [3, 4], [4, 5]], [[-2, 4], [5, -9], [0, 1]], [2.5, 6.5, 4.0]),
        ([1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1], [3.0, 2.0, 2.0, 13.0, 5.0, 5.0]),
        ([10], [1], [9.0]),
    ],
)
def test_paired_wasserstein_distances(y_true, y_pred, expected) -> None:
    from dioptra_builtins.metrics.distance import (
        _flatten_batch,
        paired_wasserstein_distances,
    )

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    expected = np.array(expected)
    result: np.ndarray = paired_wasserstein_distances(y_true, y_pred)

    assert all([isinstance(res, float) for res in result])
    assert np.allclose(expected, result)


@pytest.mark.parametrize(
    "X",
    [
        [[[1, 2, 3], [3, 4, 5]], [[-2, 4, 5], [-9, 0, 1]]],
        [[[1, 2], [3, 4], [4, 5]], [[-2, 4], [5, -9], [0, 1]]],
        [[1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1]],
        [[10], [1]],
    ],
)
def test__flatten_batch(X) -> None:
    from dioptra_builtins.metrics.distance import _flatten_batch

    X = np.array(X)

    result: np.ndarray = _flatten_batch(X)
    expected: np.ndarray = X.reshape(X.shape[0], int(np.prod(X.shape[1:])))

    assert result.shape == expected.shape
    assert np.allclose(expected, result)


@pytest.mark.parametrize(
    ("y_true", "y_pred", "order", "expected"),
    [
        ([[1, 2, 3], [3, 4, 5]], [[-2, 4, 5], [-9, 0, 1]], 1, [7.0, 20.0]),
        (
            [[1, 2], [3, 4], [4, 5]],
            [[-2, 4], [5, -9], [0, 1]],
            2,
            [3.605551275463989, 13.152946437965905, 5.656854249492381],
        ),
        (
            [[1, 2], [3, 4], [4, 5]],
            [[-2, 4], [5, -9], [0, 1]],
            np.inf,
            [3.0, 13.0, 4.0],
        ),
        ([1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1], 5, [3.0, 2.0, 2.0, 13.0, 5.0, 5.0]),
        ([1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1], 1, [3.0, 2.0, 2.0, 13.0, 5.0, 5.0]),
        (
            [1, 2, 3, 4, 5, 6],
            [-2, 4, 5, -9, 0, 1],
            np.inf,
            [3.0, 2.0, 2.0, 13.0, 5.0, 5.0],
        ),
        ([10], [1], 20, [9.0]),
    ],
)
def test__matrix_difference_l_norm(y_true, y_pred, order, expected) -> None:
    from dioptra_builtins.metrics.distance import (
        _flatten_batch,
        _matrix_difference_l_norm,
    )

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    expected = np.array(expected)
    result: np.ndarray = _matrix_difference_l_norm(y_true, y_pred, order)

    assert np.allclose(expected, result)


@pytest.mark.parametrize(
    "X",
    [
        [[[1, 2, 3], [3, 4, 5]], [[-2, 4, 5], [-9, 0, 1]]],
        [[[1, 2], [3, 4], [4, 5]], [[-2, 4], [5, -9], [0, 1]]],
        [[1, 2, 3, 4, 5, 6], [-2, 4, 5, -9, 0, 1]],
        [[10], [1]],
    ],
)
@pytest.mark.parametrize(
    "order",
    [1, 2, 3, 4, 5, 6, 8, 20, np.inf],
)
def test__normalize_batch(X, order) -> None:
    from dioptra_builtins.metrics.distance import _flatten_batch, _normalize_batch

    X = _flatten_batch(np.array(X)).astype(float)
    result: np.ndarray = _normalize_batch(X, order)

    norm = np.linalg.norm(X, axis=1, ord=order)
    norm = norm.reshape((norm.shape[0], 1))
    unnormed = result * norm
    assert np.allclose(unnormed, X)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Shared assertions for REST API unit tests."""
from typing import Any

from flask.testing import FlaskClient

from dioptra.restapi.routes import V1_ROOT

from . import actions, helpers


def assert_base_resource_contents_match_expectations(response: dict[str, Any]) -> None:
    assert isinstance(response["id"], int)
    assert isinstance(response["snapshot"], int)
    assert isinstance(response["createdOn"], str)
    assert isinstance(response["lastModifiedOn"], str)
    assert isinstance(response["latestSnapshot"], bool)
    assert isinstance(response["hasDraft"], bool)

    assert helpers.is_iso_format(response["createdOn"])
    assert helpers.is_iso_format(response["lastModifiedOn"])


def assert_user_ref_contents_matches_expectations(
    user: dict[str, Any], expected_user_id: int
) -> None:
    assert isinstance(user["id"], int)
    assert isinstance(user["username"], str)
    assert isinstance(user["url"], str)
    assert user["id"] == expected_user_id


def assert_group_ref_contents_matches_expectations(
    group: dict[str, Any], expected_group_id: int
) -> None:
    assert isinstance(group["id"], int)
    assert isinstance(group["name"], str)
    assert isinstance(group["url"], str)
    assert group["id"] == expected_group_id


def assert_tag_ref_contents_matches_expectations(tags: dict[str, Any]) -> None:
    for tag in tags:
        assert isinstance(tag["id"], int)
        assert isinstance(tag["name"], str)
        assert isinstance(tag["url"], str)


def assert_base_resource_ref_contents_matches_expectations(
    resource: dict[str, Any], expected_group_id: int
) -> None:
    assert isinstance(resource["snapshotId"], int)
    assert isinstance(resource["url"], str)
    assert_group_ref_contents_matches_expectations(
        group=resource["group"], expected_group_id=expected_group_id
    )


def assert_queue_ref_contents_matches_expectations(
    queue: dict[str, Any],
    expected_queue_id: int,
    expected_group_id: int,
) -> None:
    assert isinstance(queue["name"], str)
    assert_base_resource_ref_contents_matches_expectations(
        resource=queue, expected_group_id=expected_group_id
    )
    assert queue["snapshotId"] == expected_queue_id


def assert_experiment_ref_contents_matches_expectations(
    experiment: dict[str, Any],
    expected_experiment_id: int,
    expected_group_id: int,
) -> None:
    assert isinstance(experiment["name"], str)
    assert_base_resource_ref_contents_matches_expectations(
        resource=experiment, expected_group_id=expected_group_id
    )
    assert experiment["snapshotId"] == expected_experiment_id


def assert_entrypoint_ref_contents_matches_expectations(
    entrypoint: dict[str, Any],
    expected_entrypoint_id: int,
    expected_group_id: int,
) -> None:
    assert isinstance(entrypoint["name"], str)
    assert_base_resource_ref_contents_matches_expectations(
        resource=entrypoint, expected_group_id=expected_group_id
    )
    assert entrypoint["snapshotId"] == expected_entrypoint_id


def assert_retrieving_draft_by_resource_id_works(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving an existing resource draft by resource id works.

    Args:
        client: The Flask test client.
        resource_route: The route for the resource type.
        resource_id: The id of the resource to retrieve the draft for.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/draft", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_draft_by_id_works(
    client: FlaskClient,
    resource_route: str,
    draft_id: int,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving a draft by id works.

    Args:
        client: The Flask test client.
        resource_route: The route for the resource type.
        draft_id: The id of the draft to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{resource_route}/drafts/{draft_id}", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_retrieving_drafts_works(
    client: FlaskClient,
    resource_route: str,
    expected: list[dict[str, Any]],
    group_id: int | None = None,
    search: str | None = None,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """Assert that retrieving all drafts for a resource type works.

    Args:
        client: The Flask test client.
        resource_route: The route for the resource type.
        expected: The expected response from the API.
        group_id: The group ID used in query parameters.
        search: The search string used in query parameters.
        paging_info: The paging information used in query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    if group_id is not None:
        query_string["groupId"] = group_id

    if search is not None:
        query_string["search"] = search

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{resource_route}/drafts",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_creating_another_existing_draft_fails(
    client: FlaskClient, resource_route: str, resource_id: int
) -> None:
    """Assert that registering another draft for the same resource fails

    Args:
        client: The Flask test client.
        resource_route: The route for the resource type.
        resource_id: The id of the resource to retrieve the draft for.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = actions.create_existing_resource_draft(
        client, resource_route=resource_route, resource_id=resource_id, payload={}
    )
    assert response.status_code == 400


def assert_draft_response_contents_matches_expectations(
    response: dict[str, Any],
    expected_contents: dict[str, Any],
) -> None:
    """Assert that draft response contents is valid and matches the expected values.

    Args:
        response: The actual response from the API.
        expected_contents: The expected response from the API.
        existing_draft: If the draft is of an existing resource or not.

    Raises:
        AssertionError: If the API response does not match the expected response
            or if the response contents is not valid.
    """
    expected_keys = {
        "id",
        "group",
        "user",
        "createdOn",
        "lastModifiedOn",
        "resourceType",
        "payload",
        "resource",
        "resourceSnapshot",
        "metadata",
    }
    assert set(response.keys()) == expected_keys

    # Validate the non-Ref fields
    assert isinstance(response["id"], int)
    assert isinstance(response["createdOn"], str)
    assert isinstance(response["lastModifiedOn"], str)

    assert isinstance(response["resource"], int | None)
    assert isinstance(response["resourceSnapshot"], int | None)
    assert isinstance(response["metadata"], dict)
    assert response["resource"] == expected_contents.get("resource_id", None)
    assert response["resourceSnapshot"] == expected_contents.get(
        "resource_snapshot_id", None
    )

    assert helpers.is_iso_format(response["createdOn"])
    assert helpers.is_iso_format(response["lastModifiedOn"])

    # Validate the UserRef structure
    assert isinstance(response["user"]["id"], int)
    assert isinstance(response["user"]["username"], str)
    assert isinstance(response["user"]["url"], str)
    assert response["user"]["id"] == expected_contents["user_id"]

    # Validate the GroupRef structure
    assert isinstance(response["group"]["id"], int)
    assert isinstance(response["group"]["name"], str)
    assert isinstance(response["group"]["url"], str)
    assert response["group"]["id"] == expected_contents["group_id"]

    assert response["payload"] == expected_contents["payload"]


def assert_existing_draft_is_not_found(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
) -> None:
    """Assert that a draft of an existing resource is not found.

    Args:
        client: The Flask test client.
        resource_route: The route for the resource type.
        resource_id: The id of the resource to retrieve the draft for.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    response = client.get(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/draft",
        follow_redirects=True,
    )
    assert response.status_code == 404


def assert_new_draft_is_not_found(
    client: FlaskClient,
    resource_route: str,
    draft_id: int,
) -> None:
    """Assert that a draft of an existing resource is not found.

    Args:
        client: The Flask test client.
        resource_route: The route for the resource type.
        resource_id: The id of the resource to retrieve the draft for.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    response = client.get(
        f"/{V1_ROOT}/{resource_route}/drafts/{draft_id}",
        follow_redirects=True,
    )
    assert response.status_code == 404


def assert_retrieving_snapshots_works(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving a queue by id works.

    Args:
        client: The Flask test client.
        resource_route: The route for the resource type.
        resource_id: The id of the resource to retrieve snapshots for.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/snapshots", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_retrieving_snapshot_by_id_works(
    client: FlaskClient,
    resource_route: str,
    resource_id: int,
    snapshot_id: int,
    expected: dict[str, Any],
) -> None:
    """Assert that retrieving a resource snapshot by id works.

    Args:
        client: The Flask test client.
        resource_route: The route for the resource type.
        resource_id: The id of the resource to retrieve a snapshot of.
        snapshot_id: The id to the snapshot to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{resource_route}/{resource_id}/snapshots/{snapshot_id}",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_tags_response_contents_matches_expectations(
    response: list[dict[str, Any]],
    expected_contents: list[int],
) -> None:
    """Assert that tags response contents is valid and matches the expected values.

    Args:
        response: The actual response from the API.
        expected_contents: The expected list of Tag IDs the API.

    Raises:
        AssertionError: If the API response does not match the expected response
            or if the response contents is not valid.
    """

    # Validate the TagRef structure
    for tag in response:
        assert isinstance(tag["id"], int)
        assert isinstance(tag["name"], str)
        assert isinstance(tag["url"], str)

    assert [tag["id"] for tag in response] == expected_contents



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The schemas for serializing/deserializing Group resource."""
from __future__ import annotations

from typing import Union

from marshmallow import Schema, fields, missing

from dioptra.restapi.v1.schemas import (
    BasePageSchema,
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
)


class GroupRefSchema(Schema):
    """The reference schema for the data stored in a Group resource."""

    id = fields.Integer(
        attribute="id",
        metadata=dict(description="ID for the Group resource."),
    )
    name = fields.String(
        attribute="name", metadata=dict(description="Name of the Group resource.")
    )
    url = fields.Url(
        attribute="url",
        metadata=dict(description="URL for accessing the full Group resource."),
        relative=True,
    )


def generate_group_permissions_schema(
    is_response: bool, use_defaults: bool
) -> type[Schema]:
    schema: dict[str, Union[fields.Field, type]] = {
        "read": fields.Boolean(
            attribute="read",
            metadata=dict(description="Permission for member to read Group."),
            dump_only=is_response,
            load_only=not is_response,
            load_default=(False if use_defaults else missing),
        ),
        "write": fields.Boolean(
            attribute="write",
            metadata=dict(description="Permission for member to modify Group."),
            dump_only=is_response,
            load_only=not is_response,
            load_default=(False if use_defaults else missing),
        ),
        "shareRead": fields.Boolean(
            attribute="share_read",
            metadata=dict(
                description="Permission for member to share read-only Group."
            ),
            dump_only=is_response,
            load_only=not is_response,
            load_default=(False if use_defaults else missing),
        ),
        "shareWrite": fields.Boolean(
            attribute="share_write",
            metadata=dict(
                description="Permission for member to share read+write Group."
            ),
            dump_only=is_response,
            load_only=not is_response,
            load_default=(False if use_defaults else missing),
        ),
        "owner": fields.Boolean(
            attribute="owner",
            metadata=dict(description="Flag for if the member is a Group owner."),
            dump_only=is_response,
            load_only=not is_response,
            load_default=(False if use_defaults else missing),
        ),
        "admin": fields.Boolean(
            attribute="admin",
            metadata=dict(description="Flag for if the member is a Group admin."),
            dump_only=is_response,
            load_only=not is_response,
            load_default=(False if use_defaults else missing),
        ),
    }

    return Schema.from_dict(schema)


GroupMemberCreateFieldsSchema = generate_group_permissions_schema(
    is_response=False, use_defaults=True
)

GroupMemberMutableFieldsSchema = generate_group_permissions_schema(
    is_response=False, use_defaults=False
)

GroupPermissionsResponseSchema = generate_group_permissions_schema(
    is_response=True, use_defaults=False
)


class GroupMemberBaseSchema(Schema):
    """The base schema of a Group Member."""

    from dioptra.restapi.v1.users.schema import UserRefSchema

    userId = fields.Integer(
        attribute="user_id",
        data_key="user",
        metadata=dict(
            description="Unique identifier for the User that is a member of the Group."
        ),
        load_only=True,
        required=True,
    )
    user = fields.Nested(
        UserRefSchema,
        attribute="user",
        metadata=dict(description="User that is a member of the Group."),
        dump_only=True,
    )
    group = fields.Nested(
        GroupRefSchema,
        attribute="group",
        metadata=dict(description="The Group of which the User is a member."),
        dump_only=True,
    )
    permissions = fields.Nested(
        GroupPermissionsResponseSchema,
        attribute="permissions",
        metadata=dict(description="The Group Permissions for this User."),
        dump_only=True,
    )


class GroupMemberSchema(GroupMemberCreateFieldsSchema, GroupMemberBaseSchema):  # type: ignore
    """The schema for a Group Member"""


class GroupMutableFieldsSchema(Schema):
    """The schema for the mutable data by GroupMembers in a Group."""

    name = fields.String(
        attribute="name", metadata=dict(description="Name of the Group."), required=True
    )


class GroupSchema(GroupMutableFieldsSchema):
    """The schema for the data stored in a Group resource."""

    from dioptra.restapi.v1.users.schema import UserRefSchema

    id = fields.Integer(
        attribute="id",
        metadata=dict(description="ID for the Group resource."),
        dump_only=True,
    )
    user = fields.Nested(
        UserRefSchema,
        attribute="user",
        metadata=dict(description="User that created the Group resource."),
        dump_only=True,
    )
    members = fields.Nested(
        GroupMemberSchema,
        attribute="members",
        many=True,
        metadata=dict(description="A list of GroupMembers in a Group."),
        dump_only=True,
    )
    createdOn = fields.DateTime(
        attribute="created_on",
        metadata=dict(description="Timestamp when the Group resource was created."),
        dump_only=True,
    )
    lastModifiedOn = fields.DateTime(
        attribute="last_modified_on",
        metadata=dict(
            description="Timestamp when the Group resource was last modified."
        ),
        dump_only=True,
    )


class GroupPageSchema(BasePageSchema):
    """The paged schema for the data stored in a Group resource."""

    data = fields.Nested(
        GroupSchema,
        many=True,
        metadata=dict(description="List of Group resources in the current page."),
    )


class GroupGetQueryParameters(
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
):
    """The query parameters for the GET method of the /groups endpoint."""



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from pathlib import Path
from typing import Any, Union

import structlog
from botocore.client import BaseClient

from dioptra.sdk.utilities.paths import clear_directory
from dioptra.sdk.utilities.s3 import download_files, s3_uri_to_bucket_prefix


def _get_logger() -> Any:
    """
    Get a logger for this module.

    Returns:
        A logger object
    """
    return structlog.get_logger(__name__)


def _normalize_s3_prefix(prefix: str) -> str:
    """
    Allow a little wiggle room with S3 key prefixes, to ensure they are
    compatible with security policy.  Policy is defined with key prefix
    patterns which look like "foo/*".  This makes the slash a required
    character.  E.g. listing keys with prefix "foo" will not work, but "foo/"
    will work.

    This function ensures that a prefix which is a bare word gets a trailing
    slash.  E.g. "foo" becomes "foo/", but "foo/bar" is left alone.

    Args:
        prefix: An S3 key prefix

    Returns:
        A normalized S3 key prefix
    """
    if "/" not in prefix:
        prefix = prefix + "/"

    return prefix


def s3_download(
    s3: BaseClient,
    dest_dir: Union[str, Path],
    clear_dest: bool = False,
    preserve_key_paths: bool = False,
    *s3_uris: str,
):
    """
    Download files from S3 buckets to the local filesystem.  This function is
    intended as an implementation of a S3 CLI tool and programmatic API with
    some extra conveniences beyond the base S3 utility functions.

    Args:
        s3: A boto3 S3 client object
        dest_dir: The directory to download to; will be created if necessary
        clear_dest: If True, delete the contents of dest_dir before downloading
            anything.
        preserve_key_paths: If True, append matching keys to dest_dir to obtain
            the final directory to download to.  The intermediate
            subdirectories are created automatically.  This causes the
            directory structure represented by the keys to be mirrored in the
            local filesystem.  If False, the last component of matching key
            paths are used as filenames and the rest of the paths are
            discarded.  This causes a directory structure in the bucket to be
            flattened to list of filenames.  No subdirectories are created in
            this case.
        s3_uris: S3 URIs to download from.
    """
    log = _get_logger()

    bucket_info = []
    for s3_uri in s3_uris:
        bucket, prefix = s3_uri_to_bucket_prefix(s3_uri)

        if not bucket:
            raise ValueError("S3 URIs must include a bucket: " + s3_uri)

        prefix = _normalize_s3_prefix(prefix)

        bucket_info.append((bucket, prefix))

    # Make sure the destination dir exists!
    dest_dir_path = Path(dest_dir)
    dest_dir_path.mkdir(parents=True, exist_ok=True)

    if clear_dest:
        log.info("Clearing directory: %s", dest_dir)
        clear_directory(dest_dir_path)

    for s3_uri, (bucket, prefix) in zip(s3_uris, bucket_info):
        log.info("Downloading: %s", s3_uri)
        download_files(s3, dest_dir_path, bucket, prefix, preserve_key_paths)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import pytest

from dioptra.task_engine.type_registry import (
    TYPE_ANY,
    TYPE_BOOLEAN,
    TYPE_INTEGER,
    TYPE_NULL,
    TYPE_NUMBER,
    TYPE_STRING,
)
from dioptra.task_engine.type_validation import _infer_type, _types_compatible
from dioptra.task_engine.types import (
    SimpleType,
    StructuredType,
    StructureType,
    TypeStructure,
    UnionType,
)
from dioptra.task_engine.validation import is_valid


def test_simple_simple_compatibility() -> None:
    a_type1 = SimpleType("A")
    a_type2 = SimpleType("A")
    b_type = SimpleType("B")
    super_type = SimpleType("super")
    sub_type = SimpleType("sub", super_type=super_type)

    assert _types_compatible(a_type1, a_type2)
    assert _types_compatible(sub_type, super_type)
    assert not _types_compatible(a_type1, b_type)
    assert not _types_compatible(super_type, sub_type)


@pytest.mark.parametrize(
    "incompat_type",
    [
        StructuredType(TypeStructure(StructureType.LIST, SimpleType("A"))),
        StructuredType(
            TypeStructure(StructureType.TUPLE, [SimpleType("A"), SimpleType("B")])
        ),
        StructuredType(
            TypeStructure(
                StructureType.MAPPING, {"a": SimpleType("A"), "b": SimpleType("B")}
            )
        ),
        StructuredType(
            TypeStructure(StructureType.MAPPING, [TYPE_STRING, SimpleType("A")])
        ),
    ],
)
def test_simple_structured_compatibility(incompat_type) -> None:
    a_type = SimpleType("A")

    assert not _types_compatible(a_type, incompat_type)
    assert not _types_compatible(incompat_type, a_type)


def test_list_list_compatibility() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B", super_type=a_type)

    list_a = StructuredType(TypeStructure(StructureType.LIST, a_type))

    list_b = StructuredType(TypeStructure(StructureType.LIST, b_type))

    assert _types_compatible(list_a, list_a)
    assert _types_compatible(list_b, list_b)
    assert _types_compatible(list_b, list_a)
    assert not _types_compatible(list_a, list_b)


def test_list_tuple_compatibility() -> None:
    a_type = SimpleType("A")
    b1_type = SimpleType("B1", super_type=a_type)
    b2_type = SimpleType("B2", super_type=a_type)

    list_a = StructuredType(TypeStructure(StructureType.LIST, a_type))

    list_b1 = StructuredType(TypeStructure(StructureType.LIST, b1_type))

    tuple_b = StructuredType(TypeStructure(StructureType.TUPLE, [b1_type, b2_type]))

    assert _types_compatible(tuple_b, list_a)
    assert not _types_compatible(list_a, tuple_b)
    assert not _types_compatible(tuple_b, list_b1)
    assert not _types_compatible(list_b1, tuple_b)


def test_tuple_tuple_compatibility() -> None:
    a_type = SimpleType("A")
    b1_type = SimpleType("B1", super_type=a_type)
    b2_type = SimpleType("B2", super_type=a_type)
    c_type = SimpleType("C")

    tuple_a = StructuredType(TypeStructure(StructureType.TUPLE, [a_type]))

    tuple_a_a = StructuredType(TypeStructure(StructureType.TUPLE, [a_type, a_type]))

    tuple_b_b = StructuredType(TypeStructure(StructureType.TUPLE, [b1_type, b2_type]))

    tuple_b_b_b = StructuredType(
        TypeStructure(StructureType.TUPLE, [b1_type, b2_type, b2_type])
    )

    tuple_b_c = StructuredType(
        TypeStructure(StructureType.TUPLE, [b1_type, b2_type, c_type])
    )

    empty_tuple = StructuredType(TypeStructure(StructureType.TUPLE, []))

    assert not _types_compatible(tuple_b_b, tuple_a)
    assert not _types_compatible(tuple_a, tuple_b_b)
    assert _types_compatible(tuple_b_b, tuple_a_a)
    assert not _types_compatible(tuple_b_b, tuple_b_b_b)
    assert not _types_compatible(tuple_b_b_b, tuple_b_c)
    assert _types_compatible(tuple_b_c, tuple_b_c)
    assert _types_compatible(empty_tuple, empty_tuple)
    assert not _types_compatible(empty_tuple, tuple_a)
    assert not _types_compatible(tuple_a, empty_tuple)


def test_map_enum_map_enum_compatibility() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B", super_type=a_type)
    c_type = SimpleType("C")

    map_a = StructuredType(TypeStructure(StructureType.MAPPING, {"a": a_type}))

    map_b = StructuredType(TypeStructure(StructureType.MAPPING, {"b": b_type}))

    map_b_keyed_a = StructuredType(TypeStructure(StructureType.MAPPING, {"a": b_type}))

    map_c = StructuredType(TypeStructure(StructureType.MAPPING, {"c": c_type}))

    map_a_c = StructuredType(
        TypeStructure(StructureType.MAPPING, {"a": a_type, "c": c_type})
    )

    empty_map = StructuredType(TypeStructure(StructureType.MAPPING, {}))

    assert _types_compatible(map_a, map_a)
    assert _types_compatible(map_b_keyed_a, map_a)
    assert not _types_compatible(map_b, map_a)
    assert not _types_compatible(map_a, map_b_keyed_a)
    assert not _types_compatible(map_a, map_b)
    assert not _types_compatible(map_a, map_c)
    assert not _types_compatible(map_b, map_c)
    assert not _types_compatible(map_a, map_a_c)
    assert not _types_compatible(map_b, map_a_c)
    assert not _types_compatible(map_c, map_a_c)
    assert _types_compatible(empty_map, empty_map)
    assert not _types_compatible(empty_map, map_a)
    assert not _types_compatible(map_a, empty_map)


def test_map_kv_map_kv_compatibility() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B", super_type=a_type)
    c_type = SimpleType("C")

    map_string_a = StructuredType(
        TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type])
    )

    map_string_b = StructuredType(
        TypeStructure(StructureType.MAPPING, [TYPE_STRING, b_type])
    )

    map_string_c = StructuredType(
        TypeStructure(StructureType.MAPPING, [TYPE_STRING, c_type])
    )

    map_int_b = StructuredType(
        TypeStructure(StructureType.MAPPING, [TYPE_INTEGER, b_type])
    )

    assert _types_compatible(map_string_a, map_string_a)
    assert _types_compatible(map_string_b, map_string_a)
    assert not _types_compatible(map_string_a, map_string_b)
    assert not _types_compatible(map_string_c, map_string_a)
    assert not _types_compatible(map_int_b, map_string_b)
    assert _types_compatible(map_int_b, map_int_b)


def test_map_kv_map_enum_compatibility() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B", super_type=a_type)

    map_a = StructuredType(TypeStructure(StructureType.MAPPING, {"a": a_type}))

    map_b = StructuredType(TypeStructure(StructureType.MAPPING, {"b": b_type}))

    empty_map = StructuredType(TypeStructure(StructureType.MAPPING, {}))

    map_string_a = StructuredType(
        TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type])
    )

    map_string_b = StructuredType(
        TypeStructure(StructureType.MAPPING, [TYPE_STRING, b_type])
    )

    map_int_a = StructuredType(
        TypeStructure(StructureType.MAPPING, [TYPE_INTEGER, a_type])
    )

    assert _types_compatible(map_a, map_string_a)
    assert not _types_compatible(map_string_a, map_a)
    assert _types_compatible(map_b, map_string_a)
    assert not _types_compatible(map_string_b, map_a)
    assert not _types_compatible(map_b, map_int_a)
    assert _types_compatible(empty_map, map_string_a)
    assert _types_compatible(empty_map, map_string_b)
    assert not _types_compatible(empty_map, map_int_a)
    assert not _types_compatible(map_string_a, empty_map)
    assert not _types_compatible(map_int_a, empty_map)


def test_map_other_structured_compatibility() -> None:
    a_type = SimpleType("A")

    map_a = StructuredType(TypeStructure(StructureType.MAPPING, {"a": a_type}))

    map_string_a = StructuredType(
        TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type])
    )

    list_a = StructuredType(TypeStructure(StructureType.LIST, a_type))

    tuple_a = StructuredType(TypeStructure(StructureType.TUPLE, [a_type]))

    assert not _types_compatible(map_a, list_a)
    assert not _types_compatible(list_a, map_a)
    assert not _types_compatible(map_a, tuple_a)
    assert not _types_compatible(tuple_a, map_a)
    assert not _types_compatible(map_string_a, list_a)
    assert not _types_compatible(list_a, map_string_a)
    assert not _types_compatible(map_string_a, tuple_a)
    assert not _types_compatible(tuple_a, map_string_a)


def test_union_compatibility() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B", super_type=a_type)
    c_type = SimpleType("C")

    union_a = UnionType([a_type])
    union_b = UnionType([b_type])
    union_a_c = UnionType([a_type, c_type])

    assert _types_compatible(a_type, union_a)
    assert _types_compatible(b_type, union_a)
    assert not _types_compatible(c_type, union_a)
    assert _types_compatible(b_type, union_a_c)
    assert _types_compatible(union_a, a_type)
    assert _types_compatible(union_b, a_type)
    assert _types_compatible(union_a, union_a)
    assert _types_compatible(union_b, union_a)
    assert not _types_compatible(union_a, union_b)
    assert not _types_compatible(union_a_c, union_b)


def test_empty_union_compatibility() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")

    union_a_b = UnionType([a_type, b_type])
    empty_union = UnionType([])
    list_a = StructuredType(TypeStructure(StructureType.LIST, a_type))
    tuple_a_b = StructuredType(TypeStructure(StructureType.TUPLE, [a_type, b_type]))
    enum_map_a_b = StructuredType(
        TypeStructure(StructureType.MAPPING, {"a": a_type, "b": b_type})
    )
    kv_map_a_b = StructuredType(
        TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type])
    )

    # Treat empty union like an "empty set": it is compatible with all,
    # and nothing except itself is compatible with it.
    assert _types_compatible(empty_union, empty_union)
    assert _types_compatible(empty_union, union_a_b)
    assert _types_compatible(empty_union, a_type)
    assert _types_compatible(empty_union, list_a)
    assert _types_compatible(empty_union, tuple_a_b)
    assert _types_compatible(empty_union, enum_map_a_b)
    assert _types_compatible(empty_union, kv_map_a_b)
    assert not _types_compatible(a_type, empty_union)
    assert not _types_compatible(union_a_b, empty_union)
    assert not _types_compatible(list_a, empty_union)
    assert not _types_compatible(tuple_a_b, empty_union)
    assert not _types_compatible(enum_map_a_b, empty_union)
    assert not _types_compatible(kv_map_a_b, empty_union)


@pytest.mark.parametrize(
    "type_named, type_anon",
    [
        (
            StructuredType(
                TypeStructure(StructureType.LIST, SimpleType("A")), name="list_a"
            ),
            StructuredType(TypeStructure(StructureType.LIST, SimpleType("A"))),
        ),
        (
            StructuredType(
                TypeStructure(StructureType.TUPLE, [SimpleType("A")]), name="tuple_a"
            ),
            StructuredType(TypeStructure(StructureType.TUPLE, [SimpleType("A")])),
        ),
        (
            StructuredType(
                TypeStructure(StructureType.MAPPING, {"a": SimpleType("A")}),
                name="mapping_a_enum",
            ),
            StructuredType(
                TypeStructure(StructureType.MAPPING, {"a": SimpleType("A")})
            ),
        ),
        (
            StructuredType(
                TypeStructure(StructureType.MAPPING, [TYPE_STRING, SimpleType("A")]),
                name="mapping_a_key_value",
            ),
            StructuredType(
                TypeStructure(StructureType.MAPPING, [TYPE_STRING, SimpleType("A")])
            ),
        ),
    ],
)
def test_named_anon_type_compatibility(type_named, type_anon) -> None:
    assert _types_compatible(type_named, type_named)
    assert _types_compatible(type_anon, type_anon)
    assert _types_compatible(type_named, type_anon)
    assert _types_compatible(type_anon, type_named)


def test_named_type_compatibility() -> None:
    a_type = SimpleType("A")
    list_a1 = StructuredType(TypeStructure(StructureType.LIST, a_type), name="name1")

    list_a2 = StructuredType(TypeStructure(StructureType.LIST, a_type), name="name2")

    # list structure is compatible, but names are not equal, so the types are
    # incompatible.
    assert not _types_compatible(list_a1, list_a2)

    union_a1 = UnionType([a_type], name="name1")
    union_a2 = UnionType([a_type], name="name2")

    # ... but unions are an exception.  Their names are ignored and only their
    # members are considered.
    assert _types_compatible(union_a1, union_a2)


def test_any_compatibility() -> None:
    a_type = SimpleType("A")
    union_a_any = UnionType([a_type, TYPE_ANY])

    assert _types_compatible(a_type, TYPE_ANY)
    assert _types_compatible(TYPE_ANY, TYPE_ANY)
    assert not _types_compatible(TYPE_ANY, a_type)
    assert _types_compatible(union_a_any, union_a_any)
    assert _types_compatible(TYPE_ANY, union_a_any)
    assert _types_compatible(union_a_any, TYPE_ANY)
    assert _types_compatible(a_type, union_a_any)


@pytest.mark.parametrize(
    "value, expected_type",
    [
        (1, TYPE_INTEGER),
        (1.1, TYPE_NUMBER),
        ("foo", TYPE_STRING),
        (True, TYPE_BOOLEAN),
        (None, TYPE_NULL),
        (1 + 2j, TYPE_ANY),
    ],
)
def test_type_inference_simple(value, expected_type) -> None:
    inferred_type = _infer_type(value)
    assert inferred_type == expected_type


@pytest.mark.parametrize(
    "value, expected_type",
    [
        ([1], StructuredType(TypeStructure(StructureType.TUPLE, [TYPE_INTEGER]))),
        (
            [1, False],
            StructuredType(
                TypeStructure(StructureType.TUPLE, [TYPE_INTEGER, TYPE_BOOLEAN])
            ),
        ),
        ([], StructuredType(TypeStructure(StructureType.TUPLE, []))),
    ],
)
def test_type_inference_tuple(value, expected_type) -> None:
    inferred_type = _infer_type(value)
    assert inferred_type == expected_type


@pytest.mark.parametrize(
    "value, expected_type",
    [
        (
            {"a": 1},
            StructuredType(TypeStructure(StructureType.MAPPING, {"a": TYPE_INTEGER})),
        ),
        (
            {"a": "foo", "b": None},
            StructuredType(
                TypeStructure(StructureType.MAPPING, {"a": TYPE_STRING, "b": TYPE_NULL})
            ),
        ),
        (
            {1: "a", 2: "b"},
            StructuredType(
                TypeStructure(StructureType.MAPPING, [TYPE_INTEGER, TYPE_STRING])
            ),
        ),
        (
            {1: "a", 2: True},
            StructuredType(
                TypeStructure(
                    StructureType.MAPPING,
                    [TYPE_INTEGER, UnionType([TYPE_STRING, TYPE_BOOLEAN])],
                )
            ),
        ),
        ({1 + 2j: 1.23, None: "hello"}, TYPE_ANY),
        ({}, StructuredType(TypeStructure(StructureType.MAPPING, {}))),
    ],
)
def test_type_inference_mapping(value, expected_type) -> None:
    inferred_type = _infer_type(value)
    assert inferred_type == expected_type


def test_global_parameter_defaults_good():
    experiment_desc = {
        "parameters": {
            "int": {"type": "integer", "default": 1},
            "number": {"type": "number", "default": 1.2},
            "num_init_with_int": {"type": "number", "default": 1},
            "string": {"type": "string", "default": "abc"},
            "boolean": {"type": "boolean", "default": True},
            "null": {"type": "null", "default": None},
        },
        "tasks": {"foo": {"plugin": "org.example.foo"}},
        "graph": {"step1": {"foo": []}},
    }

    assert is_valid(experiment_desc)


@pytest.mark.parametrize(
    "bad_param",
    [
        {"type": "integer", "default": "not_an_int"},
        {"type": "integer", "default": 1.2},
        {"type": "number", "default": None},
        {"type": "null", "default": 1},
        {"type": "string", "default": 1},
        {"type": "boolean", "default": 1},
    ],
)
def test_global_parameter_defaults_bad(bad_param) -> None:
    experiment_desc = {
        "parameters": {"bad_param": bad_param},
        "tasks": {"foo": {"plugin": "org.example.foo"}},
        "graph": {"step1": {"foo": []}},
    }

    assert not is_valid(experiment_desc)


def test_positional_invocation_type_error() -> None:
    experiment_desc = {
        "tasks": {
            "task1": {"plugin": "org.example.task1", "inputs": [{"in1": "integer"}]}
        },
        "graph": {"step1": {"task1": "foo"}},
    }

    assert not is_valid(experiment_desc)


def test_positional_invocation_type_ok() -> None:
    experiment_desc = {
        "tasks": {
            "task1": {"plugin": "org.example.task1", "inputs": [{"in1": "number"}]}
        },
        "graph": {"step1": {"task1": 1}},
    }

    assert is_valid(experiment_desc)


def test_keyword_invocation_type_error() -> None:
    experiment_desc = {
        "tasks": {
            "task1": {"plugin": "org.example.task1", "inputs": [{"in1": "integer"}]}
        },
        "graph": {"step1": {"task1": {"in1": "foo"}}},
    }

    assert not is_valid(experiment_desc)


def test_keyword_invocation_type_ok() -> None:
    experiment_desc = {
        "tasks": {
            "task1": {"plugin": "org.example.task1", "inputs": [{"in1": "string"}]}
        },
        "graph": {"step1": {"task1": {"in1": "foo"}}},
    }

    assert is_valid(experiment_desc)


def test_hybrid_invocation_type_error() -> None:
    experiment_desc = {
        "tasks": {
            "task1": {
                "plugin": "org.example.task1",
                "inputs": [
                    {"in1": "integer"},
                    {"name": "in2", "type": "boolean", "required": True},
                ],
            }
        },
        "graph": {"step1": {"task": "task1", "args": "foo", "kwargs": {"in2": 1}}},
    }

    assert not is_valid(experiment_desc)


def test_hybrid_invocation_type_ok() -> None:
    experiment_desc = {
        "tasks": {
            "task1": {
                "plugin": "org.example.task1",
                "inputs": [
                    {"in1": "boolean"},
                    {"name": "in2", "type": "null", "required": True},
                ],
            }
        },
        "graph": {"step1": {"task": "task1", "args": True, "kwargs": {"in2": None}}},
    }

    assert is_valid(experiment_desc)


def test_hybrid_invocation_type_error_with_references() -> None:
    experiment_desc = {
        "tasks": {
            "task1": {
                "plugin": "org.example.task1",
                "outputs": [{"out1": "number"}, {"out2": "string"}],
            },
            "task2": {
                "plugin": "org.example.task2",
                "inputs": [{"in1": "integer"}, {"in2": "null"}],
            },
        },
        "graph": {
            "step1": {"task1": []},
            "step2": {
                "task": "task2",
                "args": "$step1.out1",
                "kwargs": {"in2": "$step1.out2"},
            },
        },
    }

    assert not is_valid(experiment_desc)


def test_hybrid_invocation_type_ok_with_references() -> None:
    experiment_desc = {
        "tasks": {
            "task1": {
                "plugin": "org.example.task1",
                "outputs": [{"out1": "integer"}, {"out2": "string"}],
            },
            "task2": {
                "plugin": "org.example.task2",
                "inputs": [{"in1": "number"}, {"in2": "any"}],
            },
        },
        "graph": {
            "step1": {"task1": []},
            "step2": {
                "task": "task2",
                "args": "$step1.out1",
                "kwargs": {"in2": "$step1.out2"},
            },
        },
    }

    assert is_valid(experiment_desc)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import Any, Dict, Optional

import structlog
from structlog.stdlib import BoundLogger

from dioptra.sdk.exceptions import TensorflowDependencyError
from dioptra.sdk.generics import fit_estimator
from dioptra.sdk.utilities.decorators import require_package

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    from tensorflow.keras import Model
    from tensorflow.keras.callbacks import History

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


@fit_estimator.register
def _(estimator: Model, x: Any, **kwargs) -> History:
    LOGGER.info(
        "Dispatch generic function",
        generic="fit_estimator",
        estimator="tensorflow.keras.Model",
        args_signature=("Model", "Any"),
    )

    history: History = fit_keras_model(
        estimator=estimator,
        x=x,
        **kwargs,
    )

    return history


@fit_estimator.register
def _(estimator: Model, x: Any, y: Any, **kwargs) -> History:
    LOGGER.info(
        "Dispatch generic function",
        generic="fit_estimator",
        estimator="tensorflow.keras.Model",
        args_signature=("Model", "Any", "Any"),
    )

    history: History = fit_keras_model(
        estimator=estimator,
        x=x,
        y=y,
        **kwargs,
    )

    return history


@require_package("tensorflow", exc_type=TensorflowDependencyError)
def fit_keras_model(
    estimator: Model,
    x: Any,
    y: Any = None,
    batch_size: Optional[int] = None,
    nb_epochs: int = 1,
    **kwargs,
) -> History:
    fit_kwargs: Dict[str, Any] = dict(
        y=y, batch_size=batch_size, epochs=nb_epochs, **kwargs
    )
    history: History = estimator.fit(
        x=x,
        **{k: v for k, v in fit_kwargs.items() if v is not None},
    )

    return history



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform queue endpoint operations."""
from __future__ import annotations

from typing import Any, Final

import structlog
from flask_login import current_user
from injector import inject
from sqlalchemy import Integer, func, select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.db.models.constants import resource_lock_types
from dioptra.restapi.errors import (
    BackendDatabaseError,
    EntityDoesNotExistError,
    EntityExistsError,
    SortParameterValidationError,
)
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.groups.service import GroupIdService
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters

LOGGER: BoundLogger = structlog.stdlib.get_logger()

RESOURCE_TYPE: Final[str] = "queue"
SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "name": lambda x: models.Queue.name.like(x, escape="/"),
    "description": lambda x: models.Queue.description.like(x, escape="/"),
    "tag": lambda x: models.Queue.tags.any(models.Tag.name.like(x, escape="/")),
}
SORTABLE_FIELDS: Final[dict[str, Any]] = {
    "name": models.Queue.name,
    "createdOn": models.Queue.created_on,
    "lastModifiedOn": models.Resource.last_modified_on,
    "description": models.Queue.description,
}


class QueueService(object):
    """The service methods for registering and managing queues by their unique id."""

    @inject
    def __init__(
        self,
        queue_name_service: QueueNameService,
        group_id_service: GroupIdService,
    ) -> None:
        """Initialize the queue service.

        All arguments are provided via dependency injection.

        Args:
            queue_name_service: A QueueNameService object.
            group_id_service: A GroupIdService object.
        """
        self._queue_name_service = queue_name_service
        self._group_id_service = group_id_service

    def create(
        self,
        name: str,
        description: str,
        group_id: int,
        commit: bool = True,
        **kwargs,
    ) -> utils.QueueDict:
        """Create a new queue.

        Args:
            name: The name of the queue. The combination of name and group_id must be
                unique.
            description: The description of the queue.
            group_id: The group that will own the queue.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created queue object.

        Raises:
            EntityExistsError: If a queue with the given name already exists.
            EntityDoesNotExistError: If the group with the provided ID does not exist.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        duplicate = self._queue_name_service.get(name, group_id=group_id, log=log)
        if duplicate is not None:
            raise EntityExistsError(
                RESOURCE_TYPE, duplicate.resource_id, name=name, group_id=group_id
            )

        group = self._group_id_service.get(group_id, error_if_not_found=True)

        resource = models.Resource(resource_type=RESOURCE_TYPE, owner=group)
        new_queue = models.Queue(
            name=name, description=description, resource=resource, creator=current_user
        )
        db.session.add(new_queue)

        if commit:
            db.session.commit()
            log.debug(
                "Queue registration successful",
                queue_id=new_queue.resource_id,
                name=new_queue.name,
            )

        return utils.QueueDict(queue=new_queue, has_draft=False)

    def get(
        self,
        group_id: int | None,
        search_string: str,
        page_index: int,
        page_length: int,
        sort_by_string: str,
        descending: bool,
        **kwargs,
    ) -> tuple[list[utils.QueueDict], int]:
        """Fetch a list of queues, optionally filtering by search string and paging
        parameters.

        Args:
            group_id: A group ID used to filter results.
            search_string: A search string used to filter results.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of queues to be returned.
            sort_by_string: The name of the column to sort.
            descending: Boolean indicating whether to sort by descending or not.

        Returns:
            A tuple containing a list of queues and the total number of queues matching
            the query.

        Raises:
            BackendDatabaseError: If the database query returns a None when counting
                the number of queues.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get full list of queues")

        filters = list()

        if group_id is not None:
            filters.append(models.Resource.group_id == group_id)

        if search_string:
            filters.append(
                construct_sql_query_filters(search_string, SEARCHABLE_FIELDS)
            )

        stmt = (
            select(func.count(models.Queue.resource_id))
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id == models.Queue.resource_snapshot_id,
            )
        )
        total_num_queues = db.session.scalars(stmt).first()

        if total_num_queues is None:
            log.error(
                "The database query returned a None when counting the number of "
                "queues when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_queues == 0:
            return [], total_num_queues

        queues_stmt = (
            select(models.Queue)
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id == models.Queue.resource_snapshot_id,
            )
            .offset(page_index)
            .limit(page_length)
        )

        if sort_by_string and sort_by_string in SORTABLE_FIELDS:
            sort_column = SORTABLE_FIELDS[sort_by_string]
            if descending:
                sort_column = sort_column.desc()
            else:
                sort_column = sort_column.asc()
            queues_stmt = queues_stmt.order_by(sort_column)
        elif sort_by_string and sort_by_string not in SORTABLE_FIELDS:
            raise SortParameterValidationError(RESOURCE_TYPE, sort_by_string)

        queues = list(db.session.scalars(queues_stmt).all())

        drafts_stmt = select(
            models.DraftResource.payload["resource_id"].as_string().cast(Integer)
        ).where(
            models.DraftResource.payload["resource_id"]
            .as_string()
            .cast(Integer)
            .in_(tuple(queue.resource_id for queue in queues)),
            models.DraftResource.user_id == current_user.user_id,
        )
        queues_dict: dict[int, utils.QueueDict] = {
            queue.resource_id: utils.QueueDict(queue=queue, has_draft=False)
            for queue in queues
        }
        for resource_id in db.session.scalars(drafts_stmt):
            queues_dict[resource_id]["has_draft"] = True

        return list(queues_dict.values()), total_num_queues


class QueueIdService(object):
    """The service methods for registering and managing queues by their unique id."""

    @inject
    def __init__(self, queue_name_service: QueueNameService) -> None:
        """Initialize the queue service.

        All arguments are provided via dependency injection.

        Args:
            queue_name_service: A QueueNameService object.
        """
        self._queue_name_service = queue_name_service

    def get(
        self,
        queue_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> utils.QueueDict | None:
        """Fetch a queue by its unique id.

        Args:
            queue_id: The unique id of the queue.
            error_if_not_found: If True, raise an error if the queue is not found.
                Defaults to False.

        Returns:
            The queue object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the queue is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get queue by id", queue_id=queue_id)

        stmt = (
            select(models.Queue)
            .join(models.Resource)
            .where(
                models.Queue.resource_id == queue_id,
                models.Queue.resource_snapshot_id == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        queue = db.session.scalars(stmt).first()

        if queue is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(RESOURCE_TYPE, queue_id=queue_id)

            return None

        drafts_stmt = (
            select(models.DraftResource.draft_resource_id)
            .where(
                models.DraftResource.payload["resource_id"].as_string().cast(Integer)
                == queue.resource_id,
                models.DraftResource.user_id == current_user.user_id,
            )
            .exists()
            .select()
        )
        has_draft = db.session.scalar(drafts_stmt)

        return utils.QueueDict(queue=queue, has_draft=has_draft)

    def modify(
        self,
        queue_id: int,
        name: str,
        description: str,
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> utils.QueueDict | None:
        """Modify a queue.

        Args:
            queue_id: The unique id of the queue.
            name: The new name of the queue.
            description: The new description of the queue.
            error_if_not_found: If True, raise an error if the group is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated queue object.

        Raises:
            EntityDoesNotExistError: If the queue is not found and `error_if_not_found`
                is True.
            EntityExistsError: If the queue name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        queue_dict = self.get(queue_id, error_if_not_found=error_if_not_found, log=log)

        if queue_dict is None:
            return None

        queue = queue_dict["queue"]
        group_id = queue.resource.group_id
        if name != queue.name:
            duplicate = self._queue_name_service.get(name, group_id=group_id, log=log)
            if duplicate is not None:
                raise EntityExistsError(
                    RESOURCE_TYPE, duplicate.resource_id, name=name, group_id=group_id
                )

        new_queue = models.Queue(
            name=name,
            description=description,
            resource=queue.resource,
            creator=current_user,
        )
        db.session.add(new_queue)

        if commit:
            db.session.commit()
            log.debug(
                "Queue modification successful",
                queue_id=queue_id,
                name=name,
                description=description,
            )

        return utils.QueueDict(queue=new_queue, has_draft=False)

    def delete(self, queue_id: int, **kwargs) -> dict[str, Any]:
        """Delete a queue.

        Args:
            queue_id: The unique id of the queue.

        Returns:
            A dictionary reporting the status of the request.

        Raises:
            EntityDoesNotExistError: If the queue is not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        stmt = select(models.Resource).filter_by(
            resource_id=queue_id, resource_type=RESOURCE_TYPE, is_deleted=False
        )
        queue_resource = db.session.scalars(stmt).first()

        if queue_resource is None:
            raise EntityDoesNotExistError(RESOURCE_TYPE, queue_id=queue_id)

        deleted_resource_lock = models.ResourceLock(
            resource_lock_type=resource_lock_types.DELETE,
            resource=queue_resource,
        )
        db.session.add(deleted_resource_lock)
        db.session.commit()
        log.debug("Queue deleted", queue_id=queue_id)

        return {"status": "Success", "queue_id": queue_id}


class QueueIdsService(object):
    """The service methods for retrieving queues from a list of ids."""

    def get(
        self,
        queue_ids: list[int],
        error_if_not_found: bool = False,
        **kwargs,
    ) -> list[models.Queue]:
        """Fetch a list of queues by their unique ids.

        Args:
            queue_ids: The unique ids of the queues.
            error_if_not_found: If True, raise an error if the queue is not found.
                Defaults to False.

        Returns:
            The queue object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the queue is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get queue by id", queue_ids=queue_ids)

        stmt = (
            select(models.Queue)
            .join(models.Resource)
            .where(
                models.Queue.resource_id.in_(tuple(queue_ids)),
                models.Queue.resource_snapshot_id == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        queues = list(db.session.scalars(stmt).all())

        if len(queues) != len(queue_ids) and error_if_not_found:
            queue_ids_missing = set(queue_ids) - set(
                queue.resource_id for queue in queues
            )
            raise EntityDoesNotExistError(
                RESOURCE_TYPE, queue_ids=list(queue_ids_missing)
            )

        return queues


class QueueNameService(object):
    """The service methods for managing queues by their name."""

    def get(
        self,
        name: str,
        group_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> models.Queue | None:
        """Fetch a queue by its name.

        Args:
            name: The name of the queue.
            group_id: The the group id of the queue.
            error_if_not_found: If True, raise an error if the queue is not found.
                Defaults to False.

        Returns:
            The queue object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the queue is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get queue by name", queue_name=name, group_id=group_id)

        stmt = (
            select(models.Queue)
            .join(models.Resource)
            .where(
                models.Queue.name == name,
                models.Resource.group_id == group_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id == models.Queue.resource_snapshot_id,
            )
        )
        queue = db.session.scalars(stmt).first()

        if queue is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(RESOURCE_TYPE, name=name)

            return None

        return queue



"""Fix primary key and unique index on PluginTaskInputParameter and PluginTaskOutputParameter

Revision ID: 5b80d059cbb4
Revises: b5cc97666e3d
Create Date: 2024-06-27 13:22:22.111192

"""

from typing import Annotated

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    MappedAsDataclass,
    mapped_column,
    sessionmaker,
)

# revision identifiers, used by Alembic.
revision = "5b80d059cbb4"
down_revision = "b5cc97666e3d"
branch_labels = None
depends_on = None


# Migration data models
intpk = Annotated[
    int,
    mapped_column(sa.BigInteger().with_variant(sa.Integer, "sqlite"), primary_key=True),
]
text_ = Annotated[str, mapped_column(sa.Text())]
bool_ = Annotated[bool, mapped_column(sa.Boolean())]


class UpgradeBase(DeclarativeBase, MappedAsDataclass):
    pass


class PluginTaskInputParameterUpgrade(UpgradeBase):
    __tablename__ = "plugin_task_input_parameters"

    plugin_file_resource_snapshot_id: Mapped[intpk]
    plugin_task_name: Mapped[text_] = mapped_column(primary_key=True)
    name: Mapped[text_] = mapped_column(nullable=False, primary_key=True)
    required: Mapped[bool_] = mapped_column(nullable=False)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    Session = sessionmaker(bind=bind)

    with op.batch_alter_table("plugin_task_input_parameters", schema=None) as batch_op:
        batch_op.add_column(sa.Column("required", sa.Boolean(), nullable=True))

    # Set the new required field on any preexisting input_parameters to True.
    with Session() as session:
        for param in session.scalars(sa.select(PluginTaskInputParameterUpgrade)):
            param.required = True

        session.commit()

    # Set the required field to nullable=False now that it is populated with values.
    with op.batch_alter_table("plugin_task_input_parameters", schema=None) as batch_op:
        batch_op.alter_column("required", nullable=False)

    with op.batch_alter_table("plugin_task_input_parameters", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_plugin_task_input_parameters"))
        batch_op.drop_index(
            batch_op.f(
                "ix_plugin_task_input_parameters_plugin_file_resource_snapshot_id"
            ),
        )
        batch_op.create_primary_key(
            batch_op.f("pk_plugin_task_input_parameters"),
            ["plugin_file_resource_snapshot_id", "plugin_task_name", "name"],
        )
        batch_op.create_index(
            batch_op.f(
                "ix_plugin_task_input_parameters_plugin_file_resource_snapshot_id"
            ),
            ["plugin_file_resource_snapshot_id", "plugin_task_name", "name"],
            unique=True,
        )

    with op.batch_alter_table("plugin_task_output_parameters", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("pk_plugin_task_output_parameters"))
        batch_op.drop_index(
            batch_op.f(
                "ix_plugin_task_output_parameters_plugin_file_resource_snapshot_id"
            )
        )
        batch_op.create_primary_key(
            batch_op.f("pk_plugin_task_output_parameters"),
            ["plugin_file_resource_snapshot_id", "plugin_task_name", "name"],
        )
        batch_op.create_index(
            batch_op.f(
                "ix_plugin_task_output_parameters_plugin_file_resource_snapshot_id"
            ),
            ["plugin_file_resource_snapshot_id", "plugin_task_name", "name"],
            unique=True,
        )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # NOTE: We are not changing the primary key and unique index back to their original
    # state because it would require dropping data from existing tables. Keeping the
    # redefinition should not cause any issues in prior versions, and redoing the
    # upgrade just drops and replaces the same index/primary key.
    with op.batch_alter_table("plugin_task_input_parameters", schema=None) as batch_op:
        batch_op.drop_column("required")

    # ### end Alembic commands ###



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform experiment endpoint operations."""
from __future__ import annotations

from typing import Any, Final, cast

import structlog
from flask_login import current_user
from injector import inject
from sqlalchemy import Integer, func, select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.db.models.constants import resource_lock_types
from dioptra.restapi.errors import (
    BackendDatabaseError,
    EntityDoesNotExistError,
    EntityExistsError,
    SortParameterValidationError,
)
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.entrypoints.service import (
    RESOURCE_TYPE as ENTRYPOINT_RESOURCE_TYPE,
)
from dioptra.restapi.v1.entrypoints.service import EntrypointIdsService
from dioptra.restapi.v1.groups.service import GroupIdService
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters

LOGGER: BoundLogger = structlog.stdlib.get_logger()

RESOURCE_TYPE: Final[str] = "experiment"
SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "name": lambda x: models.Experiment.name.like(x, escape="/"),
    "description": lambda x: models.Experiment.description.like(x, escape="/"),
    "tag": lambda x: models.Experiment.tags.any(models.Tag.name.like(x, escape="/")),
}
SORTABLE_FIELDS: Final[dict[str, Any]] = {
    "name": models.Experiment.name,
    "createdOn": models.Experiment.created_on,
    "lastModifiedOn": models.Resource.last_modified_on,
    "description": models.Experiment.description,
}


class ExperimentService(object):
    """The service methods for registering and managing experiments by their unique
    id."""

    @inject
    def __init__(
        self,
        entrypoint_ids_service: EntrypointIdsService,
        experiment_name_service: ExperimentNameService,
        group_id_service: GroupIdService,
    ) -> None:
        """Initialize the experiment service.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_ids_service: An EntrypointIdsService object.
            experiment_name_service: An ExperimentNameService object.
            group_id_service: A GroupIdService object.
        """
        self._entrypoint_ids_service = entrypoint_ids_service
        self._experiment_name_service = experiment_name_service
        self._group_id_service = group_id_service

    def create(
        self,
        name: str,
        description: str,
        group_id: int,
        entrypoint_ids: list[int] | None = None,
        commit: bool = True,
        **kwargs,
    ) -> utils.ExperimentDict:
        """Create a new experiment.

        Args:
            name: The name of the experiment. The combination of experiment and group_id
                must be unique.
            description: The description of the experiment.
            group_id: The group that will own the experiment.
            entrypoints_ids: The list of entrypoints for the experiment.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created experiment object.

        Raises:
            EntityExistsError: If an experiment with the given name already
                exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        duplicate = self._experiment_name_service.get(name, group_id=group_id, log=log)
        if duplicate is not None:
            raise EntityExistsError(
                RESOURCE_TYPE, duplicate.resource_id, name=name, group_id=group_id
            )

        group = self._group_id_service.get(group_id, error_if_not_found=True)
        entrypoints = (
            self._entrypoint_ids_service.get(
                entrypoint_ids, error_if_not_found=True, log=log
            )
            if entrypoint_ids is not None
            else []
        )

        resource = models.Resource(resource_type="experiment", owner=group)
        new_experiment = models.Experiment(
            name=name,
            description=description,
            resource=resource,
            creator=current_user,
        )
        new_experiment.children.extend(
            entrypoint.resource for entrypoint in entrypoints
        )
        db.session.add(new_experiment)

        if commit:
            db.session.commit()
            log.debug(
                "Experiment registration successful",
                experiment_id=new_experiment.resource_id,
                name=new_experiment.name,
            )

        return utils.ExperimentDict(
            experiment=new_experiment,
            entrypoints=entrypoints,
            queue=None,
            has_draft=False,
        )

    def get(
        self,
        group_id: int | None,
        search_string: str,
        page_index: int,
        page_length: int,
        sort_by_string: str,
        descending: bool,
        **kwargs,
    ) -> tuple[list[utils.ExperimentDict], int]:
        """Fetch a list of experiments, optionally filtering by search string and paging
        parameters.

        Args:
            group_id: A group ID used to filter results.
            search_string: A search string used to filter results.
            page_index: The index of the first page to be returned.
            page_length: The maximum number of experiments to be returned.
            sort_by_string: The name of the column to sort.
            descending: Boolean indicating whether to sort by descending or not.

        Returns:
            A tuple containing a list of experiments and the total number of experiments
            matching the query.

        Raises:
            BackEndDatabaseError: If the database query returns a None when counting
                the number of experiments.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get full list of experiments")

        filters = list()

        if group_id is not None:
            filters.append(models.Resource.group_id == group_id)

        if search_string:
            filters.append(
                construct_sql_query_filters(search_string, SEARCHABLE_FIELDS)
            )

        stmt = (
            select(func.count(models.Experiment.resource_id))
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.Experiment.resource_snapshot_id,
            )
        )
        total_num_experiments = db.session.scalars(stmt).first()

        if total_num_experiments is None:
            log.error(
                "The database query returned a None when counting the number of "
                "experiments when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_experiments == 0:
            return cast(list[utils.ExperimentDict], []), total_num_experiments

        experiments_stmt = (
            select(models.Experiment)
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.Experiment.resource_snapshot_id,
            )
            .offset(page_index)
            .limit(page_length)
        )

        if sort_by_string and sort_by_string in SORTABLE_FIELDS:
            sort_column = SORTABLE_FIELDS[sort_by_string]
            if descending:
                sort_column = sort_column.desc()
            else:
                sort_column = sort_column.asc()
            experiments_stmt = experiments_stmt.order_by(sort_column)
        elif sort_by_string and sort_by_string not in SORTABLE_FIELDS:
            raise SortParameterValidationError(RESOURCE_TYPE, sort_by_string)

        experiments = list(db.session.scalars(experiments_stmt).all())

        entrypoint_ids = {
            resource.resource_id: experiment.resource_id
            for experiment in experiments
            for resource in experiment.children
            if resource.resource_type == "entry_point" and not resource.is_deleted
        }
        entrypoints = self._entrypoint_ids_service.get(
            list(entrypoint_ids.keys()), error_if_not_found=True, log=log
        )
        entrypoints_dict = {
            entrypoint.resource_id: entrypoint for entrypoint in entrypoints
        }

        drafts_stmt = select(
            models.DraftResource.payload["resource_id"].as_string().cast(Integer)
        ).where(
            models.DraftResource.payload["resource_id"]
            .as_string()
            .cast(Integer)
            .in_(tuple(experiment.resource_id for experiment in experiments)),
            models.DraftResource.user_id == current_user.user_id,
        )
        experiments_dict: dict[int, utils.ExperimentDict] = {
            experiment.resource_id: utils.ExperimentDict(
                experiment=experiment, entrypoints=[], queue=None, has_draft=False
            )
            for experiment in experiments
        }
        for resource_id in db.session.scalars(drafts_stmt):
            experiments_dict[resource_id]["has_draft"] = True
        for experiment_dict in experiments_dict.values():
            for resource in experiment_dict["experiment"].children:
                if resource.resource_type == "entry_point" and not resource.is_deleted:
                    entrypoint = entrypoints_dict[resource.resource_id]
                    experiment_dict["entrypoints"].append(entrypoint)

        return list(experiments_dict.values()), total_num_experiments


class ExperimentIdService(object):
    """The service methods for registering and managing experiments by their unique
    id.
    """

    @inject
    def __init__(
        self,
        entrypoint_ids_service: EntrypointIdsService,
        experiment_name_service: ExperimentNameService,
    ) -> None:
        """Initialize the experiment service.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_ids_service: An EntrypointIdsService object.
            experiment_name_service: An ExperimentNameService object.
        """
        self._entrypoint_ids_service = entrypoint_ids_service
        self._experiment_name_service = experiment_name_service

    def get(
        self,
        experiment_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> utils.ExperimentDict | None:
        """Fetch an experiment by its unique id.

        Args:
            experiment_id: The unique id of the experiment.
            error_if_not_found: If True, raise an error if the experiment is not found.
                Defaults to False.

        Returns:
            The experiment object if found, otherwise none.

        Raises:
            EntityDoesNotExistError: If the experiment is not found and if
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get experiment by id", experiment_id=experiment_id)

        stmt = (
            select(models.Experiment)
            .join(models.Resource)
            .where(
                models.Experiment.resource_id == experiment_id,
                models.Experiment.resource_snapshot_id
                == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        experiment = db.session.scalars(stmt).first()

        if experiment is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    RESOURCE_TYPE, experiment_id=experiment_id
                )

            return None

        entrypoint_ids = [
            resource.resource_id
            for resource in experiment.children
            if resource.resource_type == "entry_point" and not resource.is_deleted
        ]
        entrypoints = self._entrypoint_ids_service.get(
            entrypoint_ids, error_if_not_found=True, log=log
        )

        drafts_stmt = (
            select(models.DraftResource.draft_resource_id)
            .where(
                models.DraftResource.payload["resource_id"].as_string().cast(Integer)
                == experiment.resource_id,
                models.DraftResource.user_id == current_user.user_id,
            )
            .exists()
            .select()
        )
        has_draft = db.session.scalar(drafts_stmt)

        return utils.ExperimentDict(
            experiment=experiment,
            entrypoints=entrypoints,
            queue=None,
            has_draft=has_draft,
        )

    def modify(
        self,
        experiment_id: int,
        name: str,
        description: str,
        entrypoint_ids: list[int],
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> utils.ExperimentDict | None:
        """Modify an experiment.

        Args:
            experiment_id: The unique if of the experiment.
            name: The new name of the experiment.
            description: The new description of the experiment.
            entrypoint_ids: The new entrypoints of the experiment.
            error_if_not_found: If True, raise an error if the experiment is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated experiment object.

        Raises:
            EntityDoesNotExistError: If the experiment is not found and
                `error_if_not_found` is True.
            EntityExistsError: If the experiment name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        experiment_dict = self.get(
            experiment_id, error_if_not_found=error_if_not_found, log=log
        )

        if experiment_dict is None:
            return None

        experiment = experiment_dict["experiment"]
        group_id = experiment.resource.group_id
        if name != experiment.name:
            duplicate = self._experiment_name_service.get(
                name, group_id=group_id, log=log
            )
            if duplicate is not None:
                raise EntityExistsError(
                    RESOURCE_TYPE, duplicate.resource_id, name=name, group_id=group_id
                )

        entrypoints = self._entrypoint_ids_service.get(
            entrypoint_ids, error_if_not_found=True, log=log
        )

        new_experiment = models.Experiment(
            name=name,
            description=description,
            resource=experiment.resource,
            creator=current_user,
        )
        new_experiment.children = [entrypoint.resource for entrypoint in entrypoints]
        db.session.add(new_experiment)

        if commit:
            db.session.commit()
            log.debug(
                "Experiment modification successful",
                experiment_id=experiment_id,
                name=name,
            )

        return utils.ExperimentDict(
            experiment=new_experiment,
            entrypoints=entrypoints,
            queue=None,
            has_draft=False,
        )

    def delete(self, experiment_id: int, **kwargs) -> dict[str, Any]:
        """Delete an experiment.

        Args:
            experiment_id: The unique id of the experiment.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        stmt = select(models.Resource).filter_by(
            resource_id=experiment_id, resource_type=RESOURCE_TYPE, is_deleted=False
        )
        experiment_resource = db.session.scalars(stmt).first()

        if experiment_resource is None:
            raise EntityDoesNotExistError(RESOURCE_TYPE, experiment_id=experiment_id)

        deleted_resource_lock = models.ResourceLock(
            resource_lock_type=resource_lock_types.DELETE,
            resource=experiment_resource,
        )
        db.session.add(deleted_resource_lock)
        db.session.commit()
        log.debug("Experiment deleted", experiment_id=experiment_id)

        return {"status": "Success", "id": [experiment_id]}


class ExperimentIdEntrypointsService(object):
    """The service methods for managing entrypoints attached to an experiment."""

    @inject
    def __init__(
        self,
        experiment_id_service: ExperimentIdService,
        entrypoint_ids_service: EntrypointIdsService,
    ) -> None:
        """Initialize the experiment service.

        All arguments are provided via dependency injection.

        Args:
            experiment_id_service: A ExperimentIdService object.
            entrypoint_ids_service: A EntrypointIdsService object.
        """
        self._experiment_id_service = experiment_id_service
        self._entrypoint_ids_service = entrypoint_ids_service

    def get(
        self,
        experiment_id: int,
        **kwargs,
    ) -> list[models.EntryPoint]:
        """Fetch the list of entrypoints for an experiment.

        Args:
            experiment_id: The unique id of the experiment.
            error_if_not_found: If True, raise an error if the experiment is not found.
                Defaults to False.

        Returns:
            The list of plugins.

        Raises:
            EntityDoesNotExistError: If the experiment is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(
            "Get entrypoints for an experiment by resource id",
            resource_id=experiment_id,
        )

        experiment_dict = cast(
            utils.ExperimentDict,
            self._experiment_id_service.get(
                experiment_id, error_if_not_found=True, log=log
            ),
        )
        return experiment_dict["entrypoints"]

    def append(
        self,
        experiment_id: int,
        entrypoint_ids: list[int],
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> list[models.EntryPoint] | None:
        """Append one or more Entrypoints to an experiment

        Args:
            experiment_id: The unique id of the experiment.
            entrypoint_ids: The list of entrypoint ids to append.
            error_if_not_found: If True, raise an error if the resource is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated list of entrypoints resource objects.

        Raises:
            EntityDoesNotExistError: If the experiment is not found and
                `error_if_not_found` is True.
            EntityDoesNotExistError: If one or more entrypoints are not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(
            "Append entrypoints to an experiment by resource id",
            resource_id=experiment_id,
        )

        experiment_dict = cast(
            utils.ExperimentDict,
            self._experiment_id_service.get(
                experiment_id, error_if_not_found=True, log=log
            ),
        )
        experiment = experiment_dict["experiment"]
        entrypoints = experiment_dict["entrypoints"]

        existing_entrypoint_ids = set(
            resource.resource_id for resource in experiment.children
        )
        new_entrypoint_ids = set(entrypoint_ids) - existing_entrypoint_ids
        new_entrypoints = self._entrypoint_ids_service.get(
            list(new_entrypoint_ids), error_if_not_found=True, log=log
        )
        if error_if_not_found and len(new_entrypoints) != len(new_entrypoint_ids):
            found_entrypoint_ids = set(
                entrypoint.resource_id for entrypoint in new_entrypoints
            )
            missing_entrypoint_ids = new_entrypoint_ids - found_entrypoint_ids
            raise EntityDoesNotExistError(
                ENTRYPOINT_RESOURCE_TYPE, entrypoint_ids=list(missing_entrypoint_ids)
            )

        experiment.children.extend(
            [entrypoint.resource for entrypoint in new_entrypoints]
        )

        if commit:
            db.session.commit()
            log.debug(
                "Entrypoints appended successfully", entrypoint_ids=entrypoint_ids
            )

        return entrypoints + new_entrypoints

    def modify(
        self,
        experiment_id: int,
        entrypoint_ids: list[int],
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> list[models.EntryPoint]:
        """Modify the list of entrypoints for an experiment.

        Args:
            experiment_id: The unique id of the experiment.
            entrypoint_ids: The list of entrypoint ids to append.
            error_if_not_found: If True, raise an error if the resource is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated entrypoint resource object.

        Raises:
            EntityDoesNotExistError: If the resource is not found and
                `error_if_not_found` is True.
            EntityDoesNotExistError: If one or more entrypoints are not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        experiment_dict = cast(
            utils.ExperimentDict,
            self._experiment_id_service.get(
                experiment_id, error_if_not_found=True, log=log
            ),
        )
        experiment = experiment_dict["experiment"]

        entrypoints = self._entrypoint_ids_service.get(
            entrypoint_ids, error_if_not_found=True, log=log
        )

        experiment.children = [entrypoint.resource for entrypoint in entrypoints]

        if commit:
            db.session.commit()
            log.debug(
                "Experiment entrypoints updated successfully",
                entrypoint_ids=entrypoint_ids,
            )

        return entrypoints

    def delete(self, experiment_id: int, **kwargs) -> dict[str, Any]:
        """Remove entrypoints from an experiment.

        Args:
            experiment_id: The unique id of the experiment.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        experiment_dict = cast(
            utils.ExperimentDict,
            self._experiment_id_service.get(
                experiment_id, error_if_not_found=True, log=log
            ),
        )
        experiment = experiment_dict["experiment"]
        experiment.children = []

        entrypoint_ids = [
            entrypoint.resource_id for entrypoint in experiment_dict["entrypoints"]
        ]

        db.session.commit()
        log.debug(
            "Entrypoints removed from experiment",
            experiment_id=experiment_id,
            entrypoint_ids=entrypoint_ids,
        )

        return {"status": "Success", "id": entrypoint_ids}


class ExperimentIdEntrypointsIdService(object):
    """The service methods for removing a entrypoint attached to an experiment."""

    @inject
    def __init__(
        self,
        experiment_id_service: ExperimentIdService,
    ) -> None:
        """Initialize the experiment service.

        All arguments are provided via dependency injection.

        Args:
            experiment_id_service: A ExperimentIdService object.
        """
        self._experiment_id_service = experiment_id_service

    def delete(self, experiment_id: int, entrypoint_id, **kwargs) -> dict[str, Any]:
        """Remove a entrypoint from an experiment.

        Args:
            experiment_id: The unique id of the experiment.
            entrypoint_id: The unique id of the entrypoint.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        experiment_dict = cast(
            utils.ExperimentDict,
            self._experiment_id_service.get(
                experiment_id, error_if_not_found=True, log=log
            ),
        )
        experiment = experiment_dict["experiment"]

        entrypoint_resources = {
            resource.resource_id: resource for resource in experiment.children
        }

        removed_entrypoint = entrypoint_resources.pop(entrypoint_id, None)

        if removed_entrypoint is None:
            raise EntityDoesNotExistError(
                ENTRYPOINT_RESOURCE_TYPE,
                experiment_id=experiment_id,
                entrypoint_id=entrypoint_id,
            )

        experiment.children = list(entrypoint_resources.values())

        db.session.commit()
        log.debug(
            "Entrypoint removed",
            experiment_id=experiment_id,
            entrypoint_id=entrypoint_id,
        )

        return {"status": "Success", "id": [entrypoint_id]}


class ExperimentNameService(object):
    """The service methods for managing experiments by their name."""

    def get(
        self,
        name: str,
        group_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> models.Experiment | None:
        """Fetch an experiment by its name.

        Args:
            name: The name of the experiment.
            group_id: The group id of the experiment.
            error_if_not_found: If True, raise an error if the experiment is not found.
                Defaults to False.

        Returns:
            The experiment object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the experiment is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get experiment by name", experiment_name=name, group_id=group_id)

        stmt = (
            select(models.Experiment)
            .join(models.Resource)
            .where(
                models.Experiment.name == name,
                models.Resource.group_id == group_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.Experiment.resource_snapshot_id,
            )
        )
        experiment = db.session.scalars(stmt).first()

        if experiment is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    RESOURCE_TYPE, experiment_name=name, group_id=group_id
                )

            return None

        return experiment



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The schemas for serializing/deserializing Plugin resources."""
import re

from marshmallow import Schema, fields, validate

from dioptra.restapi.v1.plugin_parameter_types.schema import (
    PluginParameterTypeRefSchema,
)
from dioptra.restapi.v1.schemas import (
    BasePageSchema,
    GroupIdQueryParametersSchema,
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
    generate_base_resource_ref_schema,
    generate_base_resource_schema,
)

ALLOWED_PLUGIN_NAME_REGEX = re.compile(r"^([A-Z]|[A-Z_][A-Z0-9_]+)$", flags=re.IGNORECASE)  # noqa: B950; fmt: skip
ALLOWED_PLUGIN_FILENAME_REGEX = re.compile(r"^([A-Za-z]|[A-Za-z_][A-Za-z0-9_]+)\.py$")
ALLOWED_PLUGIN_TASK_REGEX = re.compile(r"^([A-Z]|[A-Z_][A-Z0-9_]+)$", flags=re.IGNORECASE)  # noqa: B950; fmt: skip
ALLOWED_PLUGIN_TASK_PARAMETER_REGEX = re.compile(r"^([A-Z]|[A-Z_][A-Z0-9_]+)$", flags=re.IGNORECASE)  # noqa: B950; fmt: skip


PluginRefBaseSchema = generate_base_resource_ref_schema("Plugin")
PluginSnapshotRefBaseSchema = generate_base_resource_ref_schema(
    "Plugin", keep_snapshot_id=True
)


class PluginRefSchema(PluginRefBaseSchema):  # type: ignore
    """The reference schema for the data stored in a Plugin resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Plugin resource."),
    )


class PluginSnapshotRefSchema(PluginSnapshotRefBaseSchema):  # type: ignore
    """The snapshot reference schema for the data stored in a Plugin resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Plugin resource."),
    )


PluginFileRefBaseSchema = generate_base_resource_ref_schema("PluginFile")


class PluginFileRefSchema(PluginFileRefBaseSchema):  # type: ignore
    """The reference schema for the data stored in a PluginFile."""

    pluginId = fields.Int(
        attribute="plugin_id",
        data_key="plugin",
        metadata=dict(description="ID for the Plugin resource this file belongs to."),
    )
    filename = fields.String(
        attribute="filename",
        metadata=dict(description="Filename of the PluginFile resource."),
    )


class PluginTaskParameterSchema(Schema):
    """The schema for the data stored in a PluginTaskParameter"""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the PluginTaskParameter."),
        required=True,
        validate=validate.Regexp(
            ALLOWED_PLUGIN_TASK_PARAMETER_REGEX,
            error=(
                "'{input}' is not a compatible name for a Python function "
                "parameter. A Python function parameter must start with a letter or "
                "underscore, followed by letters, numbers, or underscores. In "
                "addition, '_' is not a valid Python function parameter."
            ),
        ),
    )
    parameterTypeId = fields.Int(
        attribute="parameter_type_id",
        data_key="parameterType",
        metadata=dict(
            description="The ID of the assigned PluginParameterType resource"
        ),
        load_only=True,
        required=True,
    )
    parameterType = fields.Nested(
        PluginParameterTypeRefSchema,
        attribute="parameter_type",
        metadata=dict(description="The assigned PluginParameterType resource."),
        dump_only=True,
    )


class PluginTaskInputParameterSchema(PluginTaskParameterSchema):
    """The schema for the data stored in a PluginTaskInputParameter"""

    required = fields.Boolean(
        attribute="required",
        metadata=dict(
            description=(
                "Sets whether the input parameter is required (True) or optional "
                "(False). If True, then this parameter must be assigned a value in "
                "order to execute the PluginTask."
            ),
        ),
        load_default=True,
    )


class PluginTaskOutputParameterSchema(PluginTaskParameterSchema):
    """The schema for the data stored in a PluginTaskInputParameter"""


class PluginTaskSchema(Schema):
    """The schema for the data stored in a PluginTask."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the PluginTask."),
        required=True,
        validate=validate.Regexp(
            ALLOWED_PLUGIN_TASK_REGEX,
            error=(
                "'{input}' is not a compatible name for a Python function. A Python "
                "function name must start with a letter or underscore, followed by "
                "letters, numbers, or underscores. In addition, '_' is not a valid "
                "Python function name."
            ),
        ),
    )
    inputParams = fields.Nested(
        PluginTaskInputParameterSchema,
        attribute="input_params",
        many=True,
        metadata=dict(
            description="List of input PluginTaskParameters in this PluginTask."
        ),
    )
    outputParams = fields.Nested(
        PluginTaskOutputParameterSchema,
        attribute="output_params",
        many=True,
        metadata=dict(
            description="List of output PluginTaskParameters in this PluginTask."
        ),
    )


PluginBaseSchema = generate_base_resource_schema("Plugin", snapshot=True)


class PluginMutableFieldsSchema(Schema):
    """The schema for the mutable data fields in a Plugin resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Plugin resource."),
        required=True,
        validate=validate.Regexp(
            ALLOWED_PLUGIN_NAME_REGEX,
            error=(
                "'{input}' is not a compatible name for Python module. A Python "
                "module must start with a letter or underscore, followed by letters, "
                "numbers, or underscores. In addition, a Python module name cannot "
                "be named '_' with no other characters."
            ),
        ),
    )
    description = fields.String(
        attribute="description",
        metadata=dict(description="Description of the Plugin resource."),
        load_default=None,
    )


class PluginSchema(PluginMutableFieldsSchema, PluginBaseSchema):  # type: ignore
    """The schema for the data stored in a Plugin resource."""

    files = fields.Nested(
        PluginFileRefSchema,
        attribute="files",
        metadata=dict(description="Files associated with the Plugin resource."),
        many=True,
        dump_only=True,
    )


class PluginPageSchema(BasePageSchema):
    """The paged schema for the data stored in a Plugin resource."""

    data = fields.Nested(
        PluginSchema,
        many=True,
        metadata=dict(description="List of Plugin resources in the current page."),
    )


class PluginGetQueryParameters(
    PagingQueryParametersSchema,
    GroupIdQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
):
    """The query parameters for the GET method of the /plugins endpoint."""


PluginFileBaseSchema = generate_base_resource_schema("PluginFile", snapshot=True)


class PluginFileMutableFieldsSchema(Schema):
    """The schema for the mutable data fields in a PluginFile resource."""

    filename = fields.String(
        attribute="filename",
        metadata=dict(description="Name of the PluginFile resource."),
        required=True,
        validate=validate.Regexp(
            ALLOWED_PLUGIN_FILENAME_REGEX,
            error=(
                "'{input}' is not a compatible name for a Python file. A Python "
                "filename must start with a letter or underscore, followed by letters, "
                "numbers, or underscores, and ending with a lowercase '.py'. In "
                "addition, '_.py' is not a valid Python filename."
            ),
        ),
    )
    contents = fields.String(
        attribute="contents",
        metadata=dict(description="Contents of the file."),
        required=True,
    )
    tasks = fields.Nested(
        PluginTaskSchema,
        attribute="tasks",
        metadata=dict(description="Tasks associated with the PluginFile resource."),
        many=True,
    )
    description = fields.String(
        attribute="description",
        metadata=dict(description="Description of the PluginFile resource."),
        load_default=None,
    )


class PluginFileSchema(PluginFileMutableFieldsSchema, PluginFileBaseSchema):  # type: ignore
    """The schema for the data stored in a PluginFile resource."""

    plugin = fields.Nested(
        PluginRefSchema,
        attribute="plugin",
        metadata=dict(description="The Plugin resource this file belongs to."),
        dump_only=True,
    )


class PluginFilePageSchema(BasePageSchema):
    """The paged schema for the data stored in a PluginFile resource."""

    data = fields.Nested(
        PluginFileSchema,
        many=True,
        metadata=dict(description="List of PluginFile resources in the current page."),
    )


class PluginFileGetQueryParameters(
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
):
    """The query parameters for the GET method of the /plugins/{id}/files/ endpoint."""



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The schemas for serializing/deserializing Tags."""
from __future__ import annotations

from marshmallow import Schema, fields

from dioptra.restapi.v1.groups.schema import GroupRefSchema
from dioptra.restapi.v1.schemas import (
    BasePageSchema,
    GroupIdQueryParametersSchema,
    PagingQueryParametersSchema,
    ResourceTypeQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
)
from dioptra.restapi.v1.users.schema import UserRefSchema


class TagRefSchema(Schema):
    """The reference schema for the data stored in a Tag."""

    id = fields.Integer(
        attribute="id",
        metadata=dict(description="ID for the Tag."),
        dump_only=True,
    )
    name = fields.String(
        attribute="name", metadata=dict(description="Name of the Tag.")
    )
    group = fields.Nested(
        GroupRefSchema,
        attribute="group",
        metadata=dict(description="Owner of the Tag."),
    )
    url = fields.Url(
        attribute="url",
        metadata=dict(description="URL for accessing the full Tag."),
        relative=True,
    )


class TagMutableFieldsSchema(Schema):
    """The fields schema for the mutable data in a Tag resource."""

    name = fields.String(
        attribute="name", metadata=dict(description="Name of the Tag resource.")
    )


class TagBaseSchema(Schema):
    """The reference schema for the data stored in a Tag."""

    id = fields.Integer(
        attribute="id",
        metadata=dict(description="ID for the Tag."),
        dump_only=True,
    )
    groupId = fields.Integer(
        attribute="group_id",
        data_key="group",
        metadata=dict(description="Group that owns the Tag."),
        load_only=True,
    )
    group = fields.Nested(
        GroupRefSchema,
        attribute="group",
        metadata=dict(description="Group that owns the Tag."),
        dump_only=True,
    )
    user = fields.Nested(
        UserRefSchema,
        attribute="user",
        metadata=dict(description="User that created the Tag."),
        dump_only=True,
    )
    createdOn = fields.DateTime(
        attribute="created_on",
        metadata=dict(description="Timestamp when the Tag resource was created."),
        dump_only=True,
    )
    lastModifiedOn = fields.DateTime(
        attribute="last_modified_on",
        metadata=dict(description="Timestamp when the Tag resource was last modified."),
        dump_only=True,
    )


class TagSchema(TagMutableFieldsSchema, TagBaseSchema):
    """The schema for the data stored in a Tag resource."""


class TagPageSchema(BasePageSchema):
    """The paged schema for the data stored in a Tag resource."""

    data = fields.Nested(
        TagSchema,
        many=True,
        metadata=dict(description="List of Tags in the current page."),
    )


class TagGetQueryParameters(
    PagingQueryParametersSchema,
    GroupIdQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
):
    """The query parameters for the GET method of the /tags endpoint."""


class TagResourceQueryParameters(
    PagingQueryParametersSchema, ResourceTypeQueryParametersSchema
):
    """The query parameters for the GET method of the /tags/resources/ endpoint."""



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import Protocol

import structlog
from structlog.stdlib import BoundLogger

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    import tensorflow as tf
    from tensorflow import Tensor
    from tensorflow.keras.losses import BinaryCrossentropy, CategoricalCrossentropy

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


class ComputeLabelsLossFn(Protocol):
    def __call__(
        self,
        true_labels: Tensor,
        pred_labels: Tensor,
        true_object: Tensor,
    ) -> Tensor:
        pass


class YOLOV1ClassificationLoss(object):
    def __init__(
        self,
        labels_weight: float = 1.0,
        loss_type: str = "categorical_crossentropy",
    ) -> None:
        self._labels_weight = labels_weight
        self._loss_type = loss_type
        self._compute_labels_loss = self._get_loss_fn(loss_type=loss_type)
        self._binary_crossentropy = BinaryCrossentropy(
            reduction=tf.keras.losses.Reduction.NONE
        )
        self._categorical_crossentropy = CategoricalCrossentropy(
            reduction=tf.keras.losses.Reduction.NONE
        )

    def __call__(self, *args, **kwargs) -> Tensor:
        return self.call(*args, **kwargs)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def call(
        self, pred_labels: Tensor, true_labels: Tensor, true_object: Tensor
    ) -> Tensor:
        labels_loss = self._compute_labels_loss(
            true_labels=true_labels,
            pred_labels=pred_labels,
            true_object=true_object,
        )

        return self._labels_weight * labels_loss

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _compute_original_loss(
        self,
        true_labels: Tensor,
        pred_labels: Tensor,
        true_object: Tensor,
    ) -> Tensor:
        sse_labels = tf.reduce_sum(tf.square(true_labels - pred_labels), axis=-1)
        reduction_axes = tf.range(tf.rank(sse_labels), dtype=tf.int32)[1:]
        sse_labels = tf.reduce_sum(true_object * sse_labels, axis=reduction_axes)

        # Count the number of bounding boxes in cells with an object that are
        # responsible for prediction
        num_obj_reduction_axes = tf.range(tf.rank(true_object), dtype=tf.int32)[1:]
        num_obj = tf.reduce_sum(
            tf.cast(true_object > 0, tf.float32), axis=num_obj_reduction_axes
        )

        return sse_labels / (num_obj + 1e-6)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _compute_categorical_crossentropy_loss(
        self,
        true_labels: Tensor,
        pred_labels: Tensor,
        true_object: Tensor,
    ) -> Tensor:
        cce_labels = self._categorical_crossentropy(
            y_true=true_labels, y_pred=pred_labels
        )
        reduction_axes = tf.range(tf.rank(cce_labels), dtype=tf.int32)[1:]
        cce_labels = tf.reduce_sum(true_object * cce_labels, axis=reduction_axes)

        # Count the number of bounding boxes in cells with an object that are
        # responsible for prediction
        num_obj_reduction_axes = tf.range(tf.rank(true_object), dtype=tf.int32)[1:]
        num_obj = tf.reduce_sum(
            tf.cast(true_object > 0, tf.float32), axis=num_obj_reduction_axes
        )

        return cce_labels / (num_obj + 1e-6)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _compute_binary_crossentropy_loss(
        self,
        true_labels: Tensor,
        pred_labels: Tensor,
        true_object: Tensor,
    ) -> Tensor:
        cce_labels = self._binary_crossentropy(y_true=true_labels, y_pred=pred_labels)
        reduction_axes = tf.range(tf.rank(cce_labels), dtype=tf.int32)[1:]
        cce_labels = tf.reduce_sum(true_object * cce_labels, axis=reduction_axes)

        # Count the number of bounding boxes in cells with an object that are
        # responsible for prediction
        num_obj_reduction_axes = tf.range(tf.rank(true_object), dtype=tf.int32)[1:]
        num_obj = tf.reduce_sum(
            tf.cast(true_object > 0, tf.float32), axis=num_obj_reduction_axes
        )

        return cce_labels / (num_obj + 1e-6)

    def _get_loss_fn(self, loss_type: str) -> ComputeLabelsLossFn:
        loss_fn_registry: dict[str, ComputeLabelsLossFn] = {
            "categorical_crossentropy": self._compute_categorical_crossentropy_loss,
            "binary_crossentropy": self._compute_binary_crossentropy_loss,
            "original": self._compute_original_loss,
        }
        loss_fn: ComputeLabelsLossFn | None = loss_fn_registry.get(loss_type)

        if loss_fn is None:
            raise KeyError(
                f"The loss type '{loss_type}' is not supported. The following types "
                f"are currently available: {list(loss_fn_registry.keys())}"
            )

        return loss_fn



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import os
import re
from pathlib import Path

from binaryornot.check import is_binary

PATTERN = r"{{(\s?cookiecutter)[.](.*?)}}"
RE_OBJ = re.compile(PATTERN)


# Source: https://github.com/cookiecutter/cookiecutter-django/blob/f8897bfdff9681a28bc07b3361ab51bddb71a27c/tests/test_cookiecutter_generation.py
def check_paths(paths):
    """Method to check all paths have correct substitutions."""
    # Assert that no match is found in any of the files
    for path in paths:
        if is_binary(str(path)):
            continue

        for line in path.open("r", encoding="utf-8"):
            match = RE_OBJ.search(line)
            assert match is None, f"cookiecutter variable not replaced in {path}"


# Source: https://github.com/cookiecutter/cookiecutter-django/blob/f8897bfdff9681a28bc07b3361ab51bddb71a27c/tests/test_cookiecutter_generation.py
def build_files_list(root_dir):
    """Build a list containing absolute paths to the generated files."""
    return [
        Path(dirpath) / file_path
        for dirpath, _, files in os.walk(root_dir)
        for file_path in files
    ]


def check_temp_paths_are_removed(result):
    temp_paths = [
        result.project_path / "templates",
    ]

    for temp_path in temp_paths:
        assert not temp_path.exists()


# Source: https://github.com/cookiecutter/cookiecutter-django/blob/f8897bfdff9681a28bc07b3361ab51bddb71a27c/tests/test_cookiecutter_generation.py
def test_project_generation(context, result):
    """Test that project is generated and fully rendered."""

    assert result.exit_code == 0
    assert result.exception is None
    assert result.project_path.name == context["__project_slug"]
    assert result.project_path.is_dir()

    check_temp_paths_are_removed(result)

    paths = build_files_list(result.project_path)
    assert paths
    check_paths(paths)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Common schemas for serializing/deserializing resources."""

import enum

from marshmallow import Schema, fields


def generate_base_resource_schema(name: str, snapshot: bool) -> type[Schema]:
    """Generates the base schema for a Resource."""
    from dioptra.restapi.v1.groups.schema import GroupRefSchema
    from dioptra.restapi.v1.tags.schema import TagRefSchema
    from dioptra.restapi.v1.users.schema import UserRefSchema

    schema: dict[str, fields.Field | type] = {
        "id": fields.Integer(
            attribute="id",
            metadata=dict(description=f"ID for the {name} resource."),
            dump_only=True,
        ),
        "snapshotId": fields.Integer(
            attribute="snapshot_id",
            data_key="snapshot",
            metadata=dict(description=f"ID for the underlying {name} snapshot."),
            dump_only=True,
        ),
        "groupId": fields.Integer(
            attribute="group_id",
            data_key="group",
            metadata=dict(
                description=f"ID of the Group that will own the {name} resource."
            ),
            load_only=True,
            required=True,
        ),
        "group": fields.Nested(
            GroupRefSchema,
            attribute="group",
            metadata=dict(description=f"Group that owns the {name} resource."),
            dump_only=True,
        ),
        "user": fields.Nested(
            UserRefSchema,
            attribute="user",
            metadata=dict(description=f"User that created the {name} resource."),
            dump_only=True,
        ),
        "createdOn": fields.DateTime(
            attribute="created_on",
            metadata=dict(
                description=f"Timestamp when the {name} resource was created."
            ),
            dump_only=True,
        ),
        "lastModifiedOn": fields.DateTime(
            attribute="last_modified_on",
            metadata=dict(
                description=f"Timestamp when the {name} resource was last modified."
            ),
            dump_only=True,
        ),
        "latestSnapshot": fields.Bool(
            attribute="latest_snapshot",
            metadata=dict(
                description=f"Whether or not the {name} resource is the latest version."
            ),
            dump_only=True,
        ),
        "hasDraft": fields.Bool(
            attribute="has_draft",
            metadata=dict(
                description=f"Whether a draft exists for the {name} resource."
            ),
            dump_only=True,
        ),
        "tags": fields.Nested(
            TagRefSchema,
            attribute="tags",
            metadata=dict(description=f"Tags associated with the {name} resource."),
            many=True,
            dump_only=True,
        ),
    }

    if not snapshot:
        schema.pop("snapshotId")
        schema.pop("user")
        schema.pop("lastModifiedOn")
        schema.pop("latestSnapshot")

    return Schema.from_dict(schema, name=f"{name}BaseSchema")


def generate_base_resource_ref_schema(
    name: str, keep_snapshot_id: bool = False
) -> type[Schema]:
    """Generates the base schema for a ResourceRef."""
    from dioptra.restapi.v1.groups.schema import GroupRefSchema

    schema: dict[str, fields.Field | type] = {
        "id": fields.Integer(
            attribute="id",
            metadata=dict(description=f"ID for the {name} resource."),
        ),
        "snapshotId": fields.Integer(
            attribute="snapshot_id",
            metadata=dict(description=f"Snapshot ID for the {name} resource."),
        ),
        "group": fields.Nested(
            GroupRefSchema,
            attribute="group",
            metadata=dict(description=f"Group that owns the {name} resource."),
        ),
        "url": fields.Url(
            attribute="url",
            metadata=dict(description=f"URL for accessing the full {name} resource."),
            relative=True,
        ),
    }

    if keep_snapshot_id:
        schema.pop("id")
        return Schema.from_dict(schema, name=f"{name}SnapshotRefBaseSchema")
    else:
        schema.pop("snapshotId")
        return Schema.from_dict(schema, name=f"{name}RefBaseSchema")


class BasePageSchema(Schema):
    """The base schema for adding paging to a resource endpoint."""

    index = fields.Integer(
        attribute="index",
        metadata=dict(description="Index of the current page."),
    )
    isComplete = fields.Boolean(
        attribute="is_complete",
        metadata=dict(description="Boolean indicating if more data is available."),
    )
    totalNumResults = fields.Integer(
        attribute="total_num_results",
        metadata=dict(description="Total number of results."),
    )
    first = fields.Url(
        attribute="first",
        metadata=dict(description="URL to first page in results set."),
        relative=True,
    )
    next = fields.Url(
        attribute="next",
        metadata=dict(description="URL to next page in results set."),
        relative=True,
    )
    prev = fields.Url(
        attribute="prev",
        metadata=dict(description="URL to last page in results set."),
        relative=True,
    )


class PagingQueryParametersSchema(Schema):
    """A schema for adding paging query parameters to a resource endpoint."""

    index = fields.Integer(
        attribute="index",
        metadata=dict(description="Index of the current page."),
        load_default=0,
    )
    pageLength = fields.Integer(
        attribute="page_length",
        metadata=dict(description="Number of results to return per page."),
        load_default=10,
    )


class ResourceTypeQueryParametersSchema(Schema):
    """A schema for adding resource_type query parameters to a resource endpoint."""

    resourceType = fields.String(
        attribute="resource_type",
        metadata=dict(description="Filter results by the type of resource."),
        load_default=None,
    )


class GroupIdQueryParametersSchema(Schema):
    """A schema for adding group_id query parameters to a resource endpoint."""

    groupId = fields.Integer(
        attribute="group_id",
        metadata=dict(description="Filter results by the Group ID."),
        load_default=None,
    )


class DraftTypes(enum.Enum):
    ALL = "all"
    EXISTING = "existing"
    NEW = "new"


class DraftTypeQueryParametersSchema(Schema):
    """A schema for adding draft_type query parameters to a resource endpoint."""

    draftType = fields.Enum(
        DraftTypes,
        attribute="draft_type",
        metadata=dict(
            description="The type of drafts to return: all, existing, or new."
        ),
        by_value=True,
        load_default=DraftTypes.ALL,
    )


class SearchQueryParametersSchema(Schema):
    """A schema for adding search query parameters to a resource endpoint."""

    search = fields.String(
        attribute="search",
        metadata=dict(description="Search terms for the query (* and ? wildcards)."),
        load_default="",
    )


class SortByGetQueryParametersSchema(Schema):
    """A schema for adding sort query parameters to a resource endpoint."""

    sortBy = fields.String(
        attribute="sort_by",
        metadata=dict(description="The name of the column to sort."),
        load_default="",
    )

    descending = fields.Bool(
        attribute="descending",
        metadata=dict(
            description="Boolean indicating whether to sort by descending or not."
        ),
        load_default=False,
    )


class ResourceGetQueryParameters(
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
):
    """The query parameters for the GET method of the resource endpoints."""


class IdListSchema(Schema):
    """The schema for a list of IDs."""

    ids = fields.List(
        fields.Integer(),
        attribute="ids",
        metadata=dict(description="List of identifiers for one or more objects."),
    )


class IdStatusResponseSchema(Schema):
    """A simple response for reporting a status for one or more resources."""

    id = fields.List(
        fields.Integer(),
        attribute="id",
        metadata=dict(
            description="A list of integers identifying the affected resource(s)."
        ),
    )
    status = fields.String(
        attribute="status",
        metadata=dict(description="The status of the request."),
    )


class ResourceUrlsPageSchema(BasePageSchema):
    """The paged schema for the Resource URLs."""

    data = fields.List(
        fields.String(),
        many=True,
        metadata=dict(description="List of Resource URLs in the current page."),
    )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from sqlalchemy import ForeignKeyConstraint, Index, and_, select
from sqlalchemy.orm import Mapped, column_property, mapped_column

from dioptra.restapi.db.db import bigint, intpk, text_

from .resources import Resource, ResourceSnapshot, resource_dependencies_table

# -- ORM Classes -----------------------------------------------------------------------


class Artifact(ResourceSnapshot):
    __tablename__ = "artifacts"

    # Database fields
    resource_snapshot_id: Mapped[intpk] = mapped_column(init=False)
    resource_id: Mapped[bigint] = mapped_column(init=False, nullable=False, index=True)
    uri: Mapped[text_] = mapped_column(nullable=False)

    # Derived fields (read-only)
    job_id: Mapped[bigint] = column_property(
        select(Resource.resource_id)
        .join(
            resource_dependencies_table,
            and_(
                Resource.resource_id
                == resource_dependencies_table.c.parent_resource_id,
                resource_dependencies_table.c.child_resource_id == resource_id,
            ),
        )
        .limit(1)
        .correlate_except(Resource)
        .scalar_subquery()
    )

    # Additional settings
    __table_args__ = (  # type: ignore[assignment]
        Index(None, "resource_snapshot_id", "resource_id", unique=True),
        ForeignKeyConstraint(
            ["resource_snapshot_id", "resource_id"],
            [
                "resource_snapshots.resource_snapshot_id",
                "resource_snapshots.resource_id",
            ],
        ),
    )
    __mapper_args__ = {
        "polymorphic_identity": "artifact",
    }



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
#
# This is a fork of the work
# https://github.com/gahjelle/pyplugs/blob/c75a5e1693691adc22bf91d01fad63aa67189d73/pyplugs/__init__.py  # noqa: B950
# See copyright below.
#
# Copyright (c) 2019 Geir Arne Hjelle
# Distributed under the terms of the MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""PyPlugs, decorator based plug-in architecture for Python

See {url} for more information.

Current maintainers:
--------------------

{maintainers}
"""
from __future__ import annotations

# Standard library imports
from collections import namedtuple as _namedtuple
from datetime import date as _date

from ._plugins import *  # noqa

__url__ = "https://pages.nist.gov/dioptra"


# Authors/maintainers of Pyplugs
_Author = _namedtuple("_Author", ["name", "email", "start", "end"])
_AUTHORS = [
    _Author(
        "James Glasbrenner", "jglasbrenner@mitre.org", _date(2020, 11, 24), _date.max
    ),
    _Author("Cory Miniter", "jminiter@mitre.org", _date(2020, 11, 24), _date.max),
    _Author("Howard Huang", "hhuang@mitre.org", _date(2020, 11, 24), _date.max),
    _Author("Julian Sexton", "jtsexton@mitre.org", _date(2020, 11, 24), _date.max),
    _Author("Paul Rowe", "prowe@mitre.org", _date(2020, 11, 24), _date.max),
    _Author(
        "Geir Arne Hjelle", "geirarne@gmail.com", _date(2019, 4, 1), _date(2020, 11, 23)
    ),
]

__author__ = ", ".join(a.name for a in _AUTHORS if a.start <= _date.today() < a.end)
__contact__ = ", ".join(a.email for a in _AUTHORS if a.start <= _date.today() < a.end)


# Update doc with info about maintainers
def _update_doc(doc: str) -> str:
    """Add information to doc-string

    Args:
        doc:  The doc-string to update.

    Returns:
        The updated doc-string.
    """
    # Maintainers
    maintainer_list = [
        f"+ {a.name} <{a.email}>" for a in _AUTHORS if a.start <= _date.today() < a.end
    ]
    maintainers = "\n".join(maintainer_list)

    # Add to doc-string
    return doc.format(maintainers=maintainers, url=__url__)


__doc__ = _update_doc(__doc__)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The top-level module of the Flask application.

:py:func:`~.create_app` is a factory function that instantiates a new
:py:class:`~flask.Flask` object and wires up all the configurations and third-party
dependencies.
"""
from __future__ import annotations

import os
import uuid
from typing import Any, Callable, List, Optional

import structlog
from flask import Flask, jsonify
from flask_cors import CORS
from flask_login import LoginManager
from flask_migrate import Migrate
from flask_restx import Api
from injector import Injector
from structlog.stdlib import BoundLogger

from dioptra.restapi.utils import setup_injection

from .__version__ import __version__ as DIOPTRA_VERSION
from .db import db

LOGGER: BoundLogger = structlog.stdlib.get_logger()

cors: CORS = CORS()
login_manager = LoginManager()
migrate: Migrate = Migrate()


def create_app(env: Optional[str] = None, injector: Optional[Injector] = None) -> Flask:
    """Creates and configures a fresh instance of the Dioptra REST API.

    Args:
        env: The configuration environment to use for the application. The allowed
            values are `"dev"`, `"prod"` and `"test"`. If `None`, the `"test"`
            configuration is used. The default is `None`.
        injector: A dependency injector used to invoke restx views.  If None,
            a default will be created.

    Returns:
        An initialized and configured :py:class:`~flask.Flask` object.
    """
    from .bootstrap import bind_dependencies, register_providers
    from .config import config_by_name
    from .errors import register_error_handlers
    from .routes import register_routes
    from .v1.users.service import load_user as v1_load_user

    if env is None:
        env = os.getenv("DIOPTRA_RESTAPI_ENV", "test")

    app: Flask = Flask(__name__)
    app.config.from_object(config_by_name[env])

    api: Api = Api(
        app,
        title="Dioptra REST API",
        version=DIOPTRA_VERSION,
        doc=app.config["DIOPTRA_SWAGGER_PATH"],
        url_scheme=app.config["DIOPTRA_BASE_URL"],
    )

    register_routes(api)
    register_error_handlers(api)

    login_manager.user_loader(v1_load_user)

    db.init_app(app)
    login_manager.init_app(app)

    if env != "prod":
        cors.init_app(
            app, resources={r"/api/*": {"origins": app.config["DIOPTRA_CORS_ORIGIN"]}}
        )

    with app.app_context():
        migrate.init_app(app, db, render_as_batch=True)

    @app.route("/health")
    def health():
        """An endpoint for monitoring if the REST API is responding to requests."""
        log = LOGGER.new(request_id=str(uuid.uuid4()))  # noqa: F841
        return jsonify("healthy")

    if not injector:
        modules: List[Callable[..., Any]] = [bind_dependencies]
        register_providers(modules)
        injector = Injector(modules)

    setup_injection(api, injector)

    return app
